from __future__ import annotations

from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_root.enum_desc.SpecialFunc import SpecialFunc
from argrelay_lib_server_plugin_core.plugin_delegator.DelegatorNoopBase import DelegatorNoopBase
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
)
from argrelay_schema_config_server.schema_config_server_app.CompositeNodeSchema import func_id_


class DelegatorNoopUnpluggedFunc(DelegatorNoopBase):

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        func_envelopes = [
            {
                instance_data_: {
                    func_id_: SpecialFunc.func_id_unplugged.name,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
                ReservedPropName.help_hint.name: "Function which does nothing and is not supposed to be plugged in",
                # TODO: TODO_19_67_22_89: remove `ignored_func_ids_list` - load as `FuncState.fs_unplugged`:
                # Tagged as `FuncState.fs_ignorable` but, if unplugged, it becomes `FuncState.func_id_unplugged`:
                ReservedPropName.func_state.name: FuncState.fs_ignorable.name,
                ReservedPropName.func_id.name: SpecialFunc.func_id_unplugged.name,
            },
        ]
        return func_envelopes



from __future__ import annotations

from copy import deepcopy

from argrelay_api_plugin_server_abstract.AbstractPluginServer import AbstractPluginServer
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.InterpStep import InterpStep
from argrelay_lib_root.enum_desc.PluginType import PluginType
from argrelay_schema_config_server.runtime_data_server_app.ServerConfig import ServerConfig


class AbstractInterpFactory(AbstractPluginServer):

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

        self.interp_tree_abs_paths: list[tuple[str, ...]] = []
        """
        All tree paths this plugin is plugged into.

        It is populated via `load_interp_tree_abs_paths`.
        """

        # Takes part in implementation of FS_01_89_09_24 interp tree:
        self.interp_tree_abs_paths_to_node_configs: dict[tuple[str, ...], dict] = {}
        """
        Classes derived from `AbstractInterp` (which are not direct plugins, not configured directly) are created by
        classed derived from `AbstractInterpFactory` (which are plugins with configs).
        Each `AbstractInterp` instance will require its own config called `interp_tree_node_config_dict` which are
        cloned/populated by `load_func_envelopes` and indexed by
        the abs path to the interp tree node in this `interp_tree_abs_paths_to_node_configs`.
        """

    def get_plugin_type(
        self,
    ) -> PluginType:
        return PluginType.InterpFactoryPlugin

    def load_interp_tree_abs_paths(
        self,
        this_plugin_instance_interp_tree_abs_paths: list[tuple[str, ...]],
    ):
        """
        This function is meant to be called prior to `load_func_envelopes` to
        let plugins know about all paths in the FS_01_89_09_24 `interp_tree` where they are plugged into.
        """
        pass

    def load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> tuple[
        # mapped_func_ids:
        list[str],
        # index_props:
        list[str],
        # func_data_envelopes:
        list[dict],
    ]:
        """
        Load func `data_envelope`-s taking into account `interp_tree_abs_path`.

        Takes part in implementation of FS_01_89_09_24 interp tree.

        Returns:
        *   list of mapped `func_id`-s
        *   `index_prop`-s used by these `func_id`-s
        """
        if interp_tree_abs_path in self.interp_tree_abs_paths_to_node_configs:
            raise RuntimeError(f"`{interp_tree_abs_path}` has already been loaded")
        else:
            self.interp_tree_abs_paths_to_node_configs[interp_tree_abs_path] = deepcopy(self.plugin_config_dict)

        return (
            [],
            [],
            [],
        )

    def is_root_func_loader(
        self,
    ) -> bool:
        """
        Indicates that it implements:
        *   `get_func_dynamic_index_props`
        *   `get_func_data_envelopes`
        """
        # TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor
        #       As of now, only `FirstArgInterpFactory` is supposed to return True.
        return False

    def get_func_dynamic_index_props(
        self,
    ) -> list[str]:
        """
        Provide `index_prop`-s of `index_model` for `funcs` if `is_root_func_loader` returns `True`.
        """
        # TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor
        #       As of now, it is only supposed to be implemented by `FirstArgInterpFactory`.
        raise NotImplementedError

    def get_func_data_envelopes(
        self,
    ) -> list[dict]:
        """
        Provide loaded func `data_envelope`-s if `is_root_func_loader` returns `True`.
        """
        # TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor
        #       As of now, it is only supposed to be implemented by `FirstArgInterpFactory`.
        raise NotImplementedError

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> AbstractInterp:
        raise NotImplementedError


class AbstractInterp:
    """
    `AbstractInterp` is NOT a plugin `AbstractInterpFactory` is.

    Interpreter processes command line sharing current state via :class:`InterpContext`.

    New instance of interpreter is created by (a plugin implementing) `AbstractInterpFactory` for each request.
    """

    instance_counter: int = 0
    """
    Instance counter per interp class.
    """

    def __init__(
        self,
        interp_factory_id: str,
        interp_tree_node_config_dict: dict,
        interp_ctx: InterpContext,
    ):
        self.interp_factory_id: str = interp_factory_id

        self.__class__.instance_counter += 1
        self.instance_number: int = self.instance_counter

        self.interp_tree_abs_path = list(interp_ctx.interp_tree_abs_path)

        self.interp_tree_node_config_dict: dict = interp_tree_node_config_dict
        """
        The configs for individual `AbstractInterp` are called `interp_tree_node_config_dict` and cloned/populated by
        `load_func_envelopes` based on their position in FS_01_89_09_24 interp tree.
        """

        self.interp_ctx: InterpContext = interp_ctx
        self.base_container_ipos: int = interp_ctx.curr_container_ipos

    def __str__(self) -> str:
        return f"fid: {self.interp_factory_id} path: {self.interp_ctx.interp_tree_abs_path} instance: {self.instance_number}"

    def consumes_args_at_once(self) -> bool:
        """
        Tell whether this interp consumes args one by one (FS_44_36_84_88) - see `consume_offered_args` for details.
        """
        return False

    def consume_dictated_args(self) -> bool:
        """
        Same as `consume_offered_args`, but for `dictated_arg`-s.
        """
        return False

    def consume_offered_args(self) -> bool:
        """
        Consume (usually) one arg at time: if consumed, return True, otherwise return False.

        If interp consumes all args at once,
        it has to override `consumes_args_at_once` to return `True`.

        Whether to consume more than one arg depends on whether it may
        violate FS_51_67_38_37 avoid impossible arg combinations.
        For example:
        *   FS_26_43_73_72 (func tree)
            All args (for func `data_container` or subsequent func `data_container` as arguments) has to be
            consumed one by one FS_44_36_84_88) because user is allowed to specify them in any order and
            consuming several at a time will violate FS_51_67_38_37 avoid impossible arg combinations.
        *   FS_01_89_09_24 (interp tree)
            All args can be consumed at once because user has to specify them in the order of the interp tree path.
        """
        return False

    def try_iterate(self) -> InterpStep:
        pass

    def has_fill_control(
        self,
    ) -> bool:
        """
        FS_72_53_55_13: See `DelegatorAbstract.has_fill_control` for more information.
        """
        return False

    def delegate_fill_control(
        self,
    ) -> bool:
        """
        FS_72_53_55_13: See `DelegatorAbstract.run_fill_control` for more information.
        """
        return False

    def propose_arg_completion(self) -> None:
        pass

    def next_interp(self) -> "AbstractInterp":
        """
        Return next interp factory id (or None).
        """
        return None



import json
import os

from argrelay_app_server.relay_server.UsageStatsEntry import UsageStatsEntry
from argrelay_app_server.relay_server.UsageStatsEntrySchema import usage_stats_entry_desc
from argrelay_lib_root.enum_desc.TopDir import TopDir
from argrelay_lib_root.misc_helper_common import get_argrelay_dir


class UsageStatsStore:
    """
    Stored entry for FS_87_02_77_34: usage stats.
    """

    def __init__(
        self,
    ):
        store_dir_path = os.path.join(
            get_argrelay_dir(),
            TopDir.var_dir.value,
        )
        os.makedirs(
            store_dir_path,
            exist_ok = True,
        )
        self.store_file_path: str = str(os.path.join(
            store_dir_path,
            "usage_stats",
        ))

    def store_usage_stats_entry(
        self,
        usage_stats_entry: UsageStatsEntry,
    ):
        with open(self.store_file_path, "a") as json_file:
            json.dump(
                usage_stats_entry_desc.dict_from_input_obj(usage_stats_entry),
                json_file,
            )
            json_file.write("\n")



"""
Functions to work with FS_97_64_39_94 `token_bucket`-related data.
"""

from __future__ import annotations


def token_buckets_to_token_ipos_list(
    token_buckets: list[list[int]],
) -> list[int]:
    """
    Converts `token_bucket`-s to a flat token ipos list.
    """
    token_ipos_list = []
    for token_bucket in token_buckets:
        for token_ipos in token_bucket:
            token_ipos_list.append(token_ipos)
    return token_ipos_list



from __future__ import annotations

from argrelay_schema_config_server.runtime_data_server_app.CompositeNode import (
    BaseNode,
    FuncTreeNode,
    InterpTreeNode,
    TreePathNode,
    ZeroArgNode,
)


class AbstractNodeVisitor:

    def visit_node(
        self,
        base_node: BaseNode,
    ):
        """
        It is called for the given `base_node` _before_ traversing `sub_tree`.
        """
        if isinstance(base_node, ZeroArgNode):
            self._visit_zero_arg_node(base_node)
        elif isinstance(base_node, TreePathNode):
            self._visit_tree_path_node(base_node)
        elif isinstance(base_node, InterpTreeNode):
            self._visit_interp_tree_node(base_node)
        elif isinstance(base_node, FuncTreeNode):
            self._visit_func_tree_node(base_node)
        else:
            raise NotImplementedError(base_node.node_type.name)

    def leave_node(
        self,
        base_node: BaseNode,
    ):
        """
        It is called for the given `base_node` _after_ traversing `sub_tree`.

        It can be called to recover some state which was altered for the duration of `sub_tree` traversal.
        """
        if isinstance(base_node, ZeroArgNode):
            self._leave_zero_arg_node(base_node)
        elif isinstance(base_node, TreePathNode):
            self._leave_tree_path_node(base_node)
        elif isinstance(base_node, InterpTreeNode):
            self._leave_interp_tree_node(base_node)
        elif isinstance(base_node, FuncTreeNode):
            self._leave_func_tree_node(base_node)
        else:
            raise NotImplementedError(base_node.node_type.name)

    ####################################################################################################################

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        raise NotImplementedError()

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        raise NotImplementedError()

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        raise NotImplementedError()

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        raise NotImplementedError()

    ####################################################################################################################

    def _leave_zero_arg_node(self, node: ZeroArgNode) -> None:
        pass

    def _leave_tree_path_node(self, node: TreePathNode) -> None:
        pass

    def _leave_interp_tree_node(self, node: InterpTreeNode) -> None:
        pass

    def _leave_func_tree_node(self, node: FuncTreeNode) -> None:
        pass

    ####################################################################################################################



from marshmallow import (
    fields,
    RAISE,
)

from argrelay_lib_root.enum_desc.PluginSide import PluginSide
from argrelay_lib_root.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_schema_config_server.runtime_data_server_plugin.PluginEntry import PluginEntry

plugin_enabled_ = "plugin_enabled"
plugin_side_ = "plugin_side"
plugin_config_ = "plugin_config"
plugin_module_name_ = "plugin_module_name"
plugin_class_name_ = "plugin_class_name"
plugin_dependencies_ = "plugin_dependencies"


class PluginEntrySchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = PluginEntry

    plugin_enabled = fields.Boolean(
        required = False,
        load_default = True,
    )

    plugin_side = fields.Enum(
        PluginSide,
        required = False,
        load_default = PluginSide.PluginServerSideOnly,
    )

    plugin_module_name = fields.String(
        required = True,
    )

    plugin_class_name = fields.String(
        required = True,
    )

    # List of other plugin instance ids this plugin depends on:
    plugin_dependencies = fields.List(
        fields.String(),
        required = False,
        load_default = [],
    )

    plugin_config = fields.Dict(
        required = False,
        load_default = {},
    )


_plugin_entry_example = {
    plugin_enabled_: True,
    plugin_side_: PluginSide.PluginClientSideOnly.name,
    plugin_module_name_: "SomePluginModule",
    plugin_class_name_: "SomePluginClass",
    plugin_dependencies_: [],
    plugin_config_: {
    },
}

plugin_entry_desc = TypeDesc(
    dict_schema = PluginEntrySchema(),
    ref_name = PluginEntrySchema.__name__,
    dict_example = _plugin_entry_example,
    default_file_path = "",
)



from __future__ import annotations

from marshmallow import (
    fields,
    post_load,
    pre_load,
    RAISE,
)

from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_schema_config_check_env.runtime_data_check_env_app.CheckEnvPluginConfig import CheckEnvPluginConfig
from argrelay_schema_config_server.runtime_data_server_plugin.PluginEntry import PluginEntry
from argrelay_schema_config_server.schema_config_server_plugin.BasePluginConfigSchema import (
    BasePluginConfigSchema,
    serialize_dag_to_list,
)
from argrelay_schema_config_server.schema_config_server_plugin.PluginEntrySchema import plugin_entry_desc

check_env_plugin_instances_ = "check_env_plugin_instances"


class CheckEnvPluginConfigSchema(BasePluginConfigSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = CheckEnvPluginConfig

    check_env_plugin_instances = fields.Dict(
        keys = fields.String(),
        values = fields.Nested(plugin_entry_desc.dict_schema),
        required = True,
    )

    @pre_load
    def adjust_before_validation(
        self,
        input_dict,
        **kwargs,
    ):
        # TODO: Temporary (for backward-compatibility until `check_env dry_run` gets deployed wider):
        #       allow old config schema to exist for a while:
        if check_env_plugin_instances_ not in input_dict:
            input_dict[check_env_plugin_instances_] = input_dict.pop("check_env_plugin_instance_entries")
        return input_dict

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        # Build DAG:
        check_env_plugin_instance_id_activate_order_dag = {}
        for plugin_instance_id in input_dict[check_env_plugin_instances_]:
            plugin_entry: PluginEntry = input_dict[check_env_plugin_instances_][plugin_instance_id]
            check_env_plugin_instance_id_activate_order_dag[plugin_instance_id] = plugin_entry.plugin_dependencies

        return type(self).model_class(
            **input_dict,
            check_env_plugin_instance_id_activate_list = serialize_dag_to_list(
                check_env_plugin_instance_id_activate_order_dag,
            ),
        )


check_env_plugin_config_desc = TypeDesc(
    dict_schema = CheckEnvPluginConfigSchema(),
    ref_name = CheckEnvPluginConfigSchema.__name__,
    dict_example = {
        check_env_plugin_instances_: {
            "some_check_env_plugin_instance_id": plugin_entry_desc.dict_example,
            "some_check_env_plugin_instance_id1": plugin_entry_desc.dict_example,
            "some_check_env_plugin_instance_id2": plugin_entry_desc.dict_example,
        },
    },
    default_file_path = "check_env_plugin.conf.yaml",
)



from __future__ import annotations

from copy import deepcopy
from typing import Union

from argrelay_api_plugin_server_abstract.DelegatorAbstract import DelegatorAbstract
from argrelay_api_server_cli.schema_response.AssignedValue import AssignedValue
from argrelay_app_server.composite_forest.CompositeForestExtractor import (
    extract_func_tree,
    extract_jump_tree,
)
from argrelay_app_server.composite_forest.CompositeInfoType import CompositeInfoType
from argrelay_app_server.composite_forest.DictTreeWalker import (
    DictTreeWalker,
    normalize_tree,
    sequence_starts_with,
)
from argrelay_app_server.misc_helper_server import insert_unique_to_sorted_list
from argrelay_app_server.runtime_context.AbstractArg import (
    ArgCommandIncomplete,
    ArgCommandValueDictated,
    ArgCommandValueOffered,
)
from argrelay_app_server.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay_app_server.runtime_context.InitControl import InitControl
from argrelay_app_server.runtime_context.InterpContext import (
    function_container_ipos_,
    InterpContext,
)
from argrelay_app_server.runtime_context.SearchControl import SearchControl
from argrelay_lib_root.enum_desc.CompScope import CompScope
from argrelay_lib_root.enum_desc.InterpStep import InterpStep
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_root.enum_desc.SpecialChar import SpecialChar
from argrelay_lib_root.enum_desc.ValueSource import ValueSource
from argrelay_lib_server_plugin_core.plugin_interp.AbstractInterpFactory import (
    AbstractInterp,
    AbstractInterpFactory,
)
from argrelay_lib_server_plugin_core.plugin_interp.FuncTreeInterpFactoryConfigSchema import (
    func_tree_interp_config_desc,
)
from argrelay_schema_config_server.runtime_data_server_app.EnvelopeCollection import EnvelopeCollection
from argrelay_schema_config_server.runtime_data_server_app.ServerConfig import ServerConfig
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
)
from argrelay_schema_config_server.schema_config_interp.InitControlSchema import (
    init_control_desc,
    init_types_to_values_,
)
from argrelay_schema_config_server.schema_config_interp.SearchControlSchema import (
    arg_name_to_prop_name_map_,
    populate_search_control,
    search_control_desc,
)

func_search_control_ = "func_search_control"
"""
This field is automatically populated by `FuncTreeInterpFactory` inside `interp_tree_node_config_dict`.
"""

func_init_control_ = "func_init_control"
"""
This field is automatically populated by `FuncTreeInterpFactory` inside `interp_tree_node_config_dict`.
"""

tree_step_prop_name_prefix_ = "tree_step_"
tree_step_arg_name_prefix_ = "step_"


class FuncTreeInterpFactory(AbstractInterpFactory):
    """
    Implements FS_26_43_73_72 func tree.
    Implements FS_91_88_07_23 jump tree.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

        # FS_26_43_73_72 func tree: func id to list of its absolute paths populated by `load_func_envelopes`.
        # These func tree paths are absolute within FS_33_76_82_84 composite forest.
        # Each func id can be attached to more than one leaf (hence, there is a list of paths to that func id).
        self.func_ids_to_func_abs_paths: dict[str, list[list[str]]] = {}

        # FS_91_88_07_23 jump tree
        dict_tree_walker: DictTreeWalker = DictTreeWalker(
            CompositeInfoType.jump_tree,
            extract_jump_tree(
                self.server_config.server_plugin_control.composite_forest,
            ),
        )
        self.paths_to_jump: dict[tuple[str, ...], tuple[str, ...]] = dict_tree_walker.build_paths_to_paths()
        """
        Implements FS_91_88_07_23 jump tree:
        for given `interp_tree_abs_path` (FS_01_89_09_24) selects next `interp_tree_abs_path`.
        """

        self.plugin_search_control: Union[SearchControl, None] = None

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return func_tree_interp_config_desc.dict_from_input_dict(plugin_config_dict)

    def provide_plugin_search_control(
        self,
    ) -> list[SearchControl]:
        if self.plugin_search_control:
            return [self.plugin_search_control]
        else:
            return []

    def load_interp_tree_abs_paths(
        self,
        this_plugin_instance_interp_tree_abs_paths: list[tuple[str, ...]],
    ):
        self.interp_tree_abs_paths.extend(
            this_plugin_instance_interp_tree_abs_paths,
        )

    def load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> tuple[
        # mapped_func_ids:
        list[str],
        # index_props:
        list[str],
        # func_data_envelopes:
        list[dict],
    ]:
        """
        To implement FS_26_43_73_72 func tree, this plugin loads func `data_envelope`-s automatically.

        It loops through func `data_envelope`-s and populates their tree path search props
        according to extracted `func_tree` where each function is plugged into.
        """

        (
            mapped_func_ids,
            index_props,
            func_data_envelopes,
        ) = super().load_func_envelopes(
            interp_tree_abs_path,
            func_ids_to_func_envelopes,
        )
        interp_tree_node_config_dict = self.interp_tree_abs_paths_to_node_configs[interp_tree_abs_path]

        func_selector_tree = normalize_tree(extract_func_tree(
            self.server_config.server_plugin_control.composite_forest,
            self.plugin_instance_id,
        ))

        dict_tree_walker = DictTreeWalker(
            CompositeInfoType.func_tree,
            func_selector_tree,
        )
        self.func_ids_to_func_abs_paths: dict[str, list[list[str]]] = dict_tree_walker.build_str_leaves_paths()

        # `func_envelope` (2-level map) per `func_id` per `func_abs_path`
        func_id_to_func_abs_path_to_func_envelope: dict[str, dict[tuple[str, ...], dict]] = {}
        # Loop through func `data_envelope`-s from all delegators:
        for func_id, func_envelope in func_ids_to_func_envelopes.items():
            if func_id not in self.func_ids_to_func_abs_paths:
                # TODO: TODO_19_67_22_89: remove `ignored_func_ids_list` - load as `FuncState.fs_unplugged`:
                #       This `func_id` is not plugged into given tree.
                #       It will be loaded as `FuncState.fs_unplugged`.
                continue
            func_abs_paths = self.func_ids_to_func_abs_paths[func_id]
            for func_abs_path in func_abs_paths:
                assert (
                    func_id not in func_id_to_func_abs_path_to_func_envelope
                    or
                    tuple(func_abs_path) not in func_id_to_func_abs_path_to_func_envelope[func_id]
                )
                if self._is_best_matching_abs_tree_path(func_abs_path, interp_tree_abs_path):
                    # This `func_id` is mapped into the `func_tree` under this plugin:
                    func_id_to_func_abs_path_to_func_envelope.setdefault(
                        func_id,
                        {},
                    )[tuple(func_abs_path)] = deepcopy(func_envelope)
                    mapped_func_ids.append(func_id)

        (
            extra_index_props,
            extra_func_data_envelopes,
        ) = self.populate_func_envelope_collection(
            func_id_to_func_abs_path_to_func_envelope,
            interp_tree_abs_path,
            interp_tree_node_config_dict,
        )
        index_props.extend(extra_index_props)
        func_data_envelopes.extend(extra_func_data_envelopes)

        return (
            mapped_func_ids,
            index_props,
            func_data_envelopes,
        )

    def populate_func_envelope_collection(
        self,
        func_id_to_func_abs_path_to_func_envelope: dict[str, dict[tuple[str, ...], dict]],
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict,
    ) -> tuple[
        # index_props:
        list[str],
        # func_data_envelopes:
        list[dict],
    ]:
        """
        Returns:
        *   `index_prop`-s used by all loaded funcs
        *   func `data_envelope`-s
        """

        self.populate_func_init_control(
            interp_tree_abs_path,
            interp_tree_node_config_dict,
        )
        self.populate_func_search_control(
            interp_tree_node_config_dict,
        )
        (
            interp_tree_abs_path_func_envelopes,
            prop_names,
        ) = self.populate_func_tree_props(
            func_id_to_func_abs_path_to_func_envelope,
        )
        envelope_collection = EnvelopeCollection(
            collection_name = ReservedEnvelopeClass.class_function.name,
            data_envelopes = interp_tree_abs_path_func_envelopes,
        )
        return (
            prop_names,
            envelope_collection.data_envelopes,
        )

    def _is_best_matching_abs_tree_path(
        self,
        func_abs_path,
        given_interp_tree_abs_path,
    ) -> bool:
        if sequence_starts_with(func_abs_path, given_interp_tree_abs_path):
            best_matching_abs_tree_path = given_interp_tree_abs_path
            max_matched_len = len(best_matching_abs_tree_path)
            for interp_tree_abs_path in self.interp_tree_abs_paths:
                if given_interp_tree_abs_path == interp_tree_abs_path:
                    continue
                if sequence_starts_with(func_abs_path, interp_tree_abs_path):
                    if max_matched_len < len(interp_tree_abs_path):
                        best_matching_abs_tree_path = interp_tree_abs_path
                        max_matched_len = len(best_matching_abs_tree_path)
                        break
                    elif max_matched_len == len(interp_tree_abs_path):
                        # There should not be two plugin paths leading to the same func:
                        raise RuntimeError()
                    else:
                        continue
            return given_interp_tree_abs_path == best_matching_abs_tree_path
        else:
            return False

    # noinspection PyMethodMayBeStatic
    def populate_func_init_control(
        self,
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict: dict,
    ):
        interp_tree_node_config_dict[func_init_control_] = {
            init_types_to_values_: {},
        }

        init_types_to_values = interp_tree_node_config_dict[func_init_control_][init_types_to_values_]
        for sel_ipos, path_step in enumerate(interp_tree_abs_path):
            init_types_to_values[f"{func_envelope_path_step_prop_name(sel_ipos)}"] = path_step

    def populate_func_search_control(
        self,
        interp_tree_node_config_dict,
    ):
        """
        Provide `func_search_control` based on `func_ids_to_func_abs_paths`.
        """

        plugin_search_control_dict: dict = populate_search_control(
            ReservedEnvelopeClass.class_function.name,
            {
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
            },
            [
                # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
                {"class": ReservedPropName.envelope_class.name},
            ],
        )

        # `func_search_control` should include keys from the interp tree abs path:
        arg_name_to_prop_name_map = plugin_search_control_dict[arg_name_to_prop_name_map_]

        # Include func tree path:
        max_len = max_path_len(self.func_ids_to_func_abs_paths)
        for sel_ipos in range(max_len):
            arg_name_to_prop_name_map.append({
                f"{tree_step_arg_name(sel_ipos)}": f"{func_envelope_path_step_prop_name(sel_ipos)}"
            })

        # Include other fields:
        arg_name_to_prop_name_map.append({
            "state": ReservedPropName.func_state.name
        })
        arg_name_to_prop_name_map.append({
            "id": ReservedPropName.func_id.name
        })

        interp_tree_node_config_dict[func_search_control_] = plugin_search_control_dict
        self.plugin_search_control = search_control_desc.obj_from_input_dict(plugin_search_control_dict)

    def populate_func_tree_props(
        self,
        func_envelopes_index: dict[str, dict[tuple[str, ...], dict]],
    ) -> tuple[list[dict], list[str]]:
        """
        Populates func tree properties for each func `data_envelope` based on its path in func tree.

        For each func envelope:
        *   Select its place on the func tree (provided by plugin config).
        *   Populate the path as envelope props with `tree_step_prop_name_prefix_`.
        """
        interp_tree_abs_path_func_envelopes: list[dict] = []
        prop_names: set[str] = set()
        for func_id in self.func_ids_to_func_abs_paths:
            func_abs_paths: list[list[str]] = self.func_ids_to_func_abs_paths[func_id]

            for func_abs_path in func_abs_paths:

                if (
                    func_id not in func_envelopes_index
                    or
                    tuple(func_abs_path) not in func_envelopes_index[func_id]
                ):
                    continue

                func_envelope = func_envelopes_index[func_id][tuple(func_abs_path)]

                # Include func tree path:
                for sel_ipos, path_step_id in enumerate(func_abs_path):
                    prop_name = func_envelope_path_step_prop_name(sel_ipos)
                    func_envelope[prop_name] = path_step_id
                    prop_names.add(prop_name)

                interp_tree_abs_path_func_envelopes.append(func_envelope)

        return (
            interp_tree_abs_path_func_envelopes,
            list(prop_names),
        )

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> AbstractInterp:
        return FuncTreeInterp(
            self.plugin_instance_id,
            self.interp_tree_abs_paths_to_node_configs[interp_ctx.interp_tree_abs_path],
            interp_ctx,
            self.func_ids_to_func_abs_paths,
            self.paths_to_jump,
        )


def func_envelope_path_step_prop_name(
    path_step_ipos: int,
) -> str:
    return f"{tree_step_prop_name_prefix_}{path_step_ipos}"


def tree_step_arg_name(
    path_step_ipos: int,
) -> str:
    return f"{tree_step_arg_name_prefix_}{path_step_ipos}"


def max_path_len(
    func_ids_to_func_abs_paths: dict[str, list[list[str]]],
):
    """
    Find maximum path length.
    """
    max_len: int = 0
    for func_rel_paths in func_ids_to_func_abs_paths.values():
        for func_rel_path in func_rel_paths:
            max_len = max(max_len, len(func_rel_path))
    return max_len


class FuncTreeInterp(AbstractInterp):
    """
    Implements FS_26_43_73_72 func tree.

    Finds function `data_envelope` within func tree first,
    then uses its delegator (see `DelegatorAbstract`) to find all args-related `data_envelope`-s.

    See FS_55_57_45_04 enum selector.
    """

    def __init__(
        self,
        interp_factory_id,
        interp_tree_node_config_dict: dict,
        interp_ctx: InterpContext,
        func_ids_to_func_abs_paths: dict[str, list[list[str]]],
        paths_to_jump: dict[tuple[str, ...], tuple[str, ...]],
    ):
        super().__init__(
            interp_factory_id,
            interp_tree_node_config_dict,
            interp_ctx,
        )
        self.paths_to_jump: dict[tuple[str, ...], tuple[str, ...]] = paths_to_jump

        # Allocate first container for function `data_envelope`:
        self.base_container_ipos += 1
        self.interp_ctx.envelope_containers.append(EnvelopeContainer(SearchControl()))

        self.select_next_container()
        self._apply_func_init_control()
        self._apply_func_search_control()

        self.func_ids_to_func_abs_paths: dict[str, list[list[str]]] = func_ids_to_func_abs_paths

    def _apply_func_init_control(self):
        self.interp_ctx.curr_container.assigned_prop_name_to_prop_value[
            ReservedPropName.envelope_class.name
        ] = AssignedValue(
            ReservedEnvelopeClass.class_function.name,
            ValueSource.init_value,
        )
        func_init_control: InitControl = init_control_desc.dict_schema.load(
            self.interp_tree_node_config_dict[func_init_control_],
        )
        for prop_type, prop_value in func_init_control.init_types_to_values.items():
            self.interp_ctx.curr_container.assigned_prop_name_to_prop_value[prop_type] = AssignedValue(
                prop_value,
                ValueSource.init_value,
            )
            if prop_type in self.interp_ctx.curr_container.remaining_prop_name_to_prop_value:
                del self.interp_ctx.curr_container.remaining_prop_name_to_prop_value[prop_type]

    def _apply_func_search_control(self):
        # Function `search_control` is populated based on
        # tree path (FS_01_89_09_24 interp tree + FS_26_43_73_72 func tree)
        # and plugin config (rather than data found in `data_envelope`):
        self.interp_ctx.curr_container.search_control = search_control_desc.dict_schema.load(
            self.interp_tree_node_config_dict[func_search_control_]
        )

    def _select_next_token_bucket(self) -> None:
        """
        FS_97_64_39_94: Logic for selecting `token_bucket`-s use by `EnvelopeContainer`-s.

        Each subsequent `EnvelopeContainer` uses next unused `token_bucket`
        until there is no more unused `token_bucket` left, in which case
        `EnvelopeContainer` uses `last_token_bucket_used`.
        """

        if self.interp_ctx.curr_container.used_token_bucket is None:

            if self.interp_ctx.last_token_bucket_used is None:
                self.interp_ctx.last_token_bucket_used = 0
            else:
                self.interp_ctx.last_token_bucket_used += 1
                if self.interp_ctx.last_token_bucket_used == len(self.interp_ctx.included_token_buckets):
                    self.interp_ctx.last_token_bucket_used -= 1

            self.interp_ctx.curr_container.used_token_bucket = self.interp_ctx.last_token_bucket_used

    def consume_dictated_args(self) -> bool:

        # All `dictated_arg`-s are consumed from the same `token_bucket`
        # (no need to have nested lists per `token_bucket`):
        consumed_dictated_args: list[ArgCommandValueDictated] = []
        any_consumed = False

        self._select_next_token_bucket()

        any_consumed = self.consume_one_dictated_arg_from_token_bucket(
            bucket_index = self.interp_ctx.curr_container.used_token_bucket,
            dictated_args = self.interp_ctx.remaining_dictated_args_per_bucket[
                self.interp_ctx.curr_container.used_token_bucket
            ],
            consumed_dictated_args = consumed_dictated_args,
        )

        # perform list modifications out of the prev loop:
        for dictated_arg in consumed_dictated_args:
            # If there are any consumed `dictated_arg`-s,
            # the `used_token_bucket` should be set:
            token_bucket_ipos = self.interp_ctx.curr_container.used_token_bucket
            assert token_bucket_ipos is not None
            # TODO: TODO_66_09_41_16: clarify command line processing
            #       Use helper functions which ensure consistency
            #       (like in this case, if one is removed, another is removed)
            self.interp_ctx.remaining_dictated_args_per_bucket[token_bucket_ipos].remove(dictated_arg)
            for token_ipos in dictated_arg.get_arg_tokens():
                self.interp_ctx.consumed_token_buckets[token_bucket_ipos].append(token_ipos)
                self.interp_ctx.remaining_token_buckets[token_bucket_ipos].remove(token_ipos)

        return any_consumed

    def consume_one_dictated_arg_from_token_bucket(
        self,
        bucket_index: int,
        dictated_args: list[ArgCommandValueDictated],
        consumed_dictated_args: list[ArgCommandValueDictated],
    ) -> bool:
        """
        Consumes at most one `dictated_arg` from given `bucket_index`.
        """

        any_consumed = False
        for dictated_arg in dictated_args:

            arg_name = dictated_arg.get_arg_name()
            arg_value = dictated_arg.get_arg_value()

            # TODO: FS_76_29_13_28 Why not define the order based on FS_31_70_49_15 `search_control`
            #       (instead of whatever internal order `remaining_prop_name_to_prop_value` has)?
            #       It could already be the case that `remaining_prop_name_to_prop_value` are ordered as `search_control`.
            #       Why not make it explicit?

            # See if `arg_name` matches any of `prop_name`-s:
            if arg_name in self.interp_ctx.curr_container.search_control.arg_name_to_prop_name_dict:
                prop_name = self.interp_ctx.curr_container.search_control.arg_name_to_prop_name_dict[arg_name]
                if prop_name in self.interp_ctx.curr_container.remaining_prop_name_to_prop_value:
                    self._assign_prop_name_as_explicit(
                        prop_name = prop_name,
                        arg_value = arg_value,
                        value_source = ValueSource.explicit_dictated_arg,
                    )

                    any_consumed = True
                    consumed_dictated_args.append(dictated_arg)

                    self._remove_determined_prop_name_from_remaining(prop_name)

                    # Consume one at most
                    # (caller function decides if this callee can remove one more):
                    break

        return any_consumed

    def consume_offered_args(self) -> bool:
        """
        Scans through `remaining_token_buckets` and tries to match its value against values of each type.

        Implements:
        *   FS_76_29_13_28 `command_arg` consumption priority.
        *   FS_44_36_84_88 consume args one by one:
            This func consumes all until the first remaining non-singled out arg.
        *   FS_97_64_39_94 `token_bucket`-s: consumption is limited to single bucket per `envelope_container`.
        """

        # All `offered_arg`-s are consumed from the same `token_bucket`
        # (no need to have nested lists per `token_bucket`):
        consumed_offered_args: list[ArgCommandValueOffered] = []
        any_consumed = False

        # Related to FS_51_67_38_37 avoid impossible arg combinations:
        # We can keep consuming `arg_value` without re-querying as long as matching `prop_value`-s are singled out.
        # We can also consume `arg_value` for ambiguous `prop_name` (with multiple `prop_value`-s) before re-querying.
        consumed_ambiguous_value = False

        self._select_next_token_bucket()

        any_consumed = self.consume_one_offered_arg_from_token_bucket(
            bucket_index = self.interp_ctx.curr_container.used_token_bucket,
            offered_args = self.interp_ctx.remaining_offered_args_per_bucket[
                self.interp_ctx.curr_container.used_token_bucket
            ],
            consumed_ambiguous_value = consumed_ambiguous_value,
            consumed_offered_args = consumed_offered_args,
        )

        # perform list modifications out of the prev loop:
        for offered_arg in consumed_offered_args:
            # If there are any consumed `offered_arg`-s,
            # the `used_token_bucket` should be set:
            token_bucket_ipos = self.interp_ctx.curr_container.used_token_bucket
            assert token_bucket_ipos is not None
            # TODO: TODO_66_09_41_16: clarify command line processing
            #       Use helper functions which ensure consistency
            #       (like in this case, if one is removed, another is removed)
            self.interp_ctx.remaining_offered_args_per_bucket[token_bucket_ipos].remove(offered_arg)
            self.interp_ctx.consumed_token_buckets[token_bucket_ipos].append(offered_arg.get_arg_token())
            self.interp_ctx.remaining_token_buckets[token_bucket_ipos].remove(offered_arg.get_arg_token())

        return any_consumed

    def consume_one_offered_arg_from_token_bucket(
        self,
        bucket_index: int,
        offered_args: list[ArgCommandValueOffered],
        consumed_ambiguous_value: bool,
        consumed_offered_args: list[ArgCommandValueOffered],
    ) -> bool:
        """
        Consumes at most one `offered_arg` from given `bucket_index`.
        """

        any_consumed = False
        for offered_arg in offered_args:

            arg_value = offered_arg.get_arg_value()

            # TODO: FS_76_29_13_28 Why not define the order based on FS_31_70_49_15 `search_control`
            #       (instead of whatever internal order `remaining_prop_name_to_prop_value` has)?
            #       It could already be the case that `remaining_prop_name_to_prop_value` are ordered as `search_control`.
            #       Why not make it explicit?

            # See if `arg_value` matches any of `prop_value`-s:
            for prop_name, prop_values in self.interp_ctx.curr_container.remaining_prop_name_to_prop_value.items():
                if arg_value in prop_values:
                    if (
                        len(prop_values) == 1
                        or
                        not consumed_ambiguous_value
                    ):
                        self._assign_prop_name_as_explicit(
                            prop_name = prop_name,
                            arg_value = arg_value,
                            value_source = ValueSource.explicit_offered_arg,
                        )

                        if len(prop_values) > 1:
                            # This was not a singled out `prop_value` for the given `prop_name`:
                            # allow only one ambiguous consumption without re-querying
                            # (for the `prop_name` according to the priority set by `search_control`).
                            # If more than one `prop_name` has multiple `prop_value`-s,
                            # some of the combinations of (`prop_name`, `prop_value`) pairs may not match
                            # the data and cannot be assigned, but it is possible to assign the first such `prop_name`
                            # with multiple (ambiguous) `prop_value` before re-querying without violating
                            # FS_51_67_38_37 avoid impossible arg combinations.
                            consumed_ambiguous_value = True

                        any_consumed = True
                        consumed_offered_args.append(offered_arg)

                        self._remove_determined_prop_name_from_remaining(prop_name)

                        # Consume one at most
                        # (caller function decides if this callee can remove one more):
                        break

        return any_consumed

    def _remove_determined_prop_name_from_remaining(
        self,
        prop_name: str,
    ):
        # TD_76_09_29_31: overlapped
        # Assign matching `prop_value` to the first `prop_name` it matches only once =
        # remove that `prop_name` from remaining:
        del self.interp_ctx.curr_container.remaining_prop_name_to_prop_value[prop_name]

    def _assign_prop_name_as_explicit(
        self,
        prop_name: str,
        arg_value: str,
        value_source: ValueSource,
    ):
        # Related to FS_13_51_07_97 singled out implicit values:
        # If `prop_name` is singled out (assigned as `ValueSource.implicit_value`)
        # but `offered_arg` or `dictated_arg` still determines `prop_name`,
        # `prop_name` must be updated as assigned by one of the explicit `ValueSource.*`.
        assert (
            value_source is ValueSource.explicit_dictated_arg
            or
            value_source is ValueSource.explicit_offered_arg
        )
        self.interp_ctx.curr_container.assigned_prop_name_to_prop_value[prop_name] = AssignedValue(
            arg_value,
            value_source,
        )

    def try_iterate(self) -> InterpStep:
        """
        Try to consume more args if possible.

        *   If function was found, start with search for its first envelope class.
        *   If curr envelope class is found, move to the next (until all are found).

        :returns:
        *   `InterpStep.NextInterp`: move to next interpreter: curr interpreter is fully satisfied from the args
        *   `InterpStep.NextEnvelope`: call again curr interpreter: still more things to find in the args
        *   `InterpStep.StopAll`: interpreter sees no point to continue the loop (`InterpContext.interpret_command`)
        """

        # We want single `data_envelope` to be found, not zero, not more than one:
        if self.interp_ctx.curr_container.found_count > 1:
            # Too many `data_envelope`-s - stop:
            return InterpStep.StopAll
        elif self.interp_ctx.curr_container.found_count == 1:

            # Func container or beyond:
            if self.interp_ctx.curr_container_ipos >= self.base_container_ipos:
                # Func param envelopes beyond func envelope: 0 for the first func param:
                func_param_container_offset = self.interp_ctx.curr_container_ipos - self.base_container_ipos
                # TODO: optimize it:
                #       *   we request one `search_control` from a list of them (internally)
                #       *   then, we give this item to allocate `searchable_container` but it may have already been allocated before:
                search_control: SearchControl = self.get_search_control(func_param_container_offset)
                while search_control is not None:
                    # Create `EnvelopeContainer`-s for next `data_envelope`-s to find:
                    self.interp_ctx.alloc_searchable_container(
                        self.base_container_ipos,
                        func_param_container_offset,
                        search_control,
                    )
                    func_param_container_offset += 1
                    search_control: SearchControl = self.get_search_control(func_param_container_offset)

            if self.interp_ctx.is_last_container():
                # Function does not need any envelopes:
                return InterpStep.NextInterp
            else:
                self.select_next_container()
                self.delegate_init_control()
                # Need more args to consume for the next envelope to find:
                return InterpStep.NextEnvelope

        else:
            # No `data_envelope` = nothing to do:
            return InterpStep.StopAll

    def get_search_control(
        self,
        func_param_container_offset: int,
    ) -> Union[SearchControl, None]:
        delegator_plugin = self.get_func_delegator()

        search_control: SearchControl = delegator_plugin.run_search_control(
            self.interp_ctx,
            self.get_found_func_data_envelope(),
            func_param_container_offset,
        )
        return search_control

    def delegate_init_control(self):
        delegator_plugin = self.get_func_delegator()
        delegator_plugin.run_init_control(
            self.interp_ctx.envelope_containers,
            self.interp_ctx.curr_container_ipos,
        )

    def has_fill_control(
        self,
    ) -> bool:
        return True

    def delegate_fill_control(
        self,
    ) -> bool:
        delegator_plugin = self.get_func_delegator()
        if delegator_plugin:
            return delegator_plugin.run_fill_control(
                self.interp_ctx,
            )
        else:
            return False

    def get_found_func_data_envelope(self) -> Union[dict, None]:
        func_envelope = self.interp_ctx.envelope_containers[(
            # TODO: TODO_73_23_85_93: use helper to select container ipos:
            self.base_container_ipos + function_container_ipos_
        )]
        if func_envelope.found_count == 1:
            return func_envelope.data_envelopes[0]
        return None

    def get_func_delegator(self):
        func_data_envelope = self.get_found_func_data_envelope()
        if func_data_envelope:
            delegator_plugin_instance_id = func_data_envelope[instance_data_][delegator_plugin_instance_id_]
            delegator_plugin: DelegatorAbstract = self.interp_ctx.action_delegators[delegator_plugin_instance_id]
            return delegator_plugin
        else:
            # func envelope hasn't been found yet:
            return None

    def select_next_container(self):
        self.interp_ctx.curr_container_ipos += 1
        self.interp_ctx.curr_container = self.interp_ctx.envelope_containers[
            self.interp_ctx.curr_container_ipos
        ]

    def next_interp(self) -> "AbstractInterp":
        delegator_plugin = self.get_func_delegator()
        interp_factory_id = delegator_plugin.run_interp_control(self)
        if interp_factory_id:
            self.select_next_interp_tree_abs_path()
            return self.interp_ctx.create_next_interp(interp_factory_id)
        else:
            return None

    def select_next_interp_tree_abs_path(self):
        if self.interp_ctx.interp_tree_abs_path in self.paths_to_jump:
            # FS_91_88_07_23 jump tree: replace current `interp_tree_abs_path` with another one based on config:
            self.interp_ctx.interp_tree_abs_path = self.paths_to_jump[self.interp_ctx.interp_tree_abs_path]

    def propose_arg_completion(self) -> None:
        comp_values = self.propose_auto_comp_list()
        for comp_value in comp_values:
            insert_unique_to_sorted_list(self.interp_ctx.comp_suggestions, comp_value)

    def propose_auto_comp_list(self) -> list[str]:

        # TODO: TODO_51_14_50_19: ensure `tangent_token` always exists
        #       This may simplify this condition:
        if self.interp_ctx.parsed_ctx.tan_token_ipos >= 0:

            if self.interp_ctx.parsed_ctx.tan_token_ipos in self.interp_ctx.token_ipos_to_arg_map:
                tangent_arg = self.interp_ctx.token_ipos_to_arg_map[self.interp_ctx.parsed_ctx.tan_token_ipos]
                if isinstance(tangent_arg, ArgCommandIncomplete):
                    incomplete_arg: ArgCommandIncomplete = tangent_arg
                    return self._propose_for_incomplete_arg_with_tangent_token(incomplete_arg)
                elif isinstance(tangent_arg, ArgCommandValueDictated):
                    dictated_arg: ArgCommandValueDictated = tangent_arg
                    return self._propose_for_dictated_arg(dictated_arg)
                elif isinstance(tangent_arg, ArgCommandValueOffered):
                    offered_arg: ArgCommandValueOffered = tangent_arg
                    return self._propose_for_offered_arg(offered_arg)
                else:
                    raise TypeError(f"unhandled argument type: {type(tangent_arg)}")
            else:
                # TODO: TODO_66_09_41_16: clarify command line processing
                #       It makes sense to model every token as `command_arg` including `token_bucket` delimiter
                #       instead of using `excluded_tokens`, no?
                # This must be `token_bucket` delimiter.
                assert self.interp_ctx.parsed_ctx.tan_token_value == SpecialChar.TokenBucketDelimiter.value
                # TODO: TODO_66_09_41_16: clarify command line processing
                #       `token_bucket` delimiter should be suggested everywhere where applicable
                #       especially when `tan_token_l_part` starts with `SpecialChar.TokenBucketDelimiter.value`.

        elif self.interp_ctx.parsed_ctx.prev_token_ipos >= 0:

            if self.interp_ctx.parsed_ctx.prev_token_ipos in self.interp_ctx.token_ipos_to_arg_map:
                prev_arg = self.interp_ctx.token_ipos_to_arg_map[self.interp_ctx.parsed_ctx.prev_token_ipos]
                if isinstance(prev_arg, ArgCommandIncomplete):
                    incomplete_arg: ArgCommandIncomplete = prev_arg
                    return self._propose_for_incomplete_arg_with_prev_token(incomplete_arg)
                else:
                    return self._propose_remaining_prop_values_based_on_comp_scope()

        else:
            return self._propose_remaining_prop_values_based_on_comp_scope()

    def _propose_for_incomplete_arg_with_tangent_token(
        self,
        incomplete_arg: ArgCommandIncomplete,
    ):
        prop_name_token_ipos = incomplete_arg.get_arg_tokens()[0]
        assert self.interp_ctx.parsed_ctx.tan_token_ipos == prop_name_token_ipos
        return self._propose_arg_names_based_on_comp_scope()

    def _propose_for_incomplete_arg_with_prev_token(
        self,
        incomplete_arg: ArgCommandIncomplete,
    ):
        prop_name_token_ipos = incomplete_arg.get_arg_tokens()[0]
        assert self.interp_ctx.parsed_ctx.prev_token_ipos == prop_name_token_ipos
        arg_name = incomplete_arg.get_arg_name()
        if arg_name in self.interp_ctx.curr_container.search_control.arg_name_to_prop_name_dict:
            prop_name = self.interp_ctx.curr_container.search_control.arg_name_to_prop_name_dict[arg_name]
            return self._remaining_prop_values_from_given_prop_name(prop_name)
        else:
            return []

    def _propose_for_dictated_arg(
        self,
        dictated_arg: ArgCommandValueDictated,
    ):
        prop_name_token_ipos = dictated_arg.get_arg_tokens()[0]
        prop_value_token_ipos = dictated_arg.get_arg_tokens()[1]
        if self.interp_ctx.parsed_ctx.tan_token_ipos == prop_name_token_ipos:
            return self._propose_arg_names_based_on_comp_scope()
        elif self.interp_ctx.parsed_ctx.tan_token_ipos == prop_value_token_ipos:
            arg_name = dictated_arg.get_arg_name()
            if arg_name in self.interp_ctx.curr_container.search_control.arg_name_to_prop_name_dict:
                prop_name = self.interp_ctx.curr_container.search_control.arg_name_to_prop_name_dict[arg_name]
                return self._remaining_prop_values_from_given_prop_name(prop_name)
            else:
                return []
        else:
            raise ValueError("unexpected")

    def _propose_for_offered_arg(
        self,
        offered_arg: ArgCommandValueOffered,
    ):
        prop_value_token_ipos = offered_arg.get_arg_token()
        assert (
            self.interp_ctx.parsed_ctx.tan_token_ipos == prop_value_token_ipos
            or
            self.interp_ctx.parsed_ctx.prev_token_ipos == prop_value_token_ipos
        )
        return self._propose_remaining_prop_values_based_on_comp_scope()

    def _propose_arg_names_based_on_comp_scope(self) -> list[str]:
        """
        Complete `arg_name` (as `arg_name` 1-to-1 corresponds to `prop_name`):
        *   If initial completion, complete `arg_name` from remaining `prop_name`-s.
        *   If subsequent completion, complete `arg_name` from all `prop_name`-s of current `search_control`.
        """

        arg_name_prefix = self.interp_ctx.parsed_ctx.tan_token_l_part[1:]
        # TODO: TODO_66_09_41_16: clarify command line processing
        #       Selected `envelope_container` should depend on which
        #       `token_bucket` this `tangent_token` belongs to.
        #       There can be many `envelope_container`-s using that `token_bucket` -
        #       we may need to list suggestions from all (not from one).
        selected_container = self.interp_ctx.curr_container

        if self.interp_ctx.parsed_ctx.comp_scope in [
            CompScope.ScopeInitial,
            CompScope.ScopeUnknown,
        ]:
            prefixed_arg_names = []
            for prop_name in selected_container.remaining_prop_name_to_prop_value:
                arg_name = selected_container.search_control.prop_name_to_arg_name_dict[prop_name]
                if (
                    # TODO: TODO_66_09_41_16: clarify command line processing
                    #       What does this char mean? Put it into one of the `SpecialChar.*`.
                    not arg_name.startswith("_")
                    and
                    arg_name.startswith(arg_name_prefix)
                ):
                    prefixed_arg_names.append(SpecialChar.ArgNamePrefix.value + arg_name)
            return prefixed_arg_names

        if self.interp_ctx.parsed_ctx.comp_scope is CompScope.ScopeSubsequent:
            prefixed_arg_names = []
            for arg_name in selected_container.search_control.prop_name_to_arg_name_dict.values():
                if (
                    # TODO: TODO_66_09_41_16: clarify command line processing
                    #       What does this char mean? Put it into one of the `SpecialChar.*`.
                    not arg_name.startswith("_")
                    and
                    arg_name.startswith(arg_name_prefix)
                ):
                    prefixed_arg_names.append(SpecialChar.ArgNamePrefix.value + arg_name)
            return prefixed_arg_names

        return []

    def _propose_remaining_prop_values_based_on_comp_scope(self) -> list[str]:
        # TODO: FS_23_62_89_43: `tangent_token`
        #       The logic for both if-s (`if-A` and `if-B`) is identical at the moment - what do we want to improve?

        # TODO: FS_23_62_89_43: if-A:
        if self.interp_ctx.parsed_ctx.comp_scope in [
            CompScope.ScopeInitial,
            CompScope.ScopeUnknown,
        ]:
            if self.interp_ctx.parsed_ctx.tan_token_l_part == "":
                return self._remaining_prop_values_from_next_missing_prop_name()
            else:
                return self._remaining_prop_values_from_next_missing_prop_name()

        # TODO: FS_23_62_89_43: if-B:
        if self.interp_ctx.parsed_ctx.comp_scope is CompScope.ScopeSubsequent:
            if self.interp_ctx.parsed_ctx.tan_token_l_part == "":
                return self._remaining_prop_values_from_next_missing_prop_name()
            else:
                # TODO: FS_20_88_05_60: Suggest `arg_name`-s (:) for missing `prop_name`-s instead - it is `ScopeSubsequent`, user insist and wants something else:
                return self._remaining_prop_values_from_next_missing_prop_name()

        return []

    def _remaining_prop_values_from_next_missing_prop_name(self) -> list[str]:
        """
        Clarifications:
        *   remaining = because values for the given type are reduced based on narrowed down `data_envelope` set
        *   missing = because this `prop_name` is not specified yet
        *   next = because `prop_name`-s are tired in specific order
        """
        proposed_values: list[str] = []

        # Return filtered value set from the next missing arg:
        for prop_name in self.interp_ctx.curr_container.search_control.prop_name_to_arg_name_dict:
            if prop_name in self.interp_ctx.curr_container.remaining_prop_name_to_prop_value:
                # TODO: TODO_66_09_41_16: clarify command line processing
                #       only one condition should be enough:
                #       `prop_name` is either in assigned or in remaining, not in both:
                assert prop_name not in self.interp_ctx.curr_container.assigned_prop_name_to_prop_value

                proposed_values = [
                    # FS_71_87_33_52: `help_hint`:
                    self.interp_ctx.help_hint_cache.get_value_with_help_hint(prop_name, x)
                    for x in self.interp_ctx.curr_container.remaining_prop_name_to_prop_value[prop_name]
                    if (
                        # TODO: TODO_66_09_41_16: clarify command line processing
                        #       This does not make sense - only str are expected.
                        #       Clean this up (or, at least, keep assert for a while).
                        # NOTE: This checks for `str` meaning primitive (scalar, not `list` or `dict`),
                        #       but this also filters out `int` or others.
                        #       Primitive types should be converted to `str` when loaded.
                        #       See `ConfigOnlyLoader.convert_envelope_fields_to_string` for example.
                        isinstance(x, str)
                        and
                        # FS_32_05_46_00: using `startswith`:
                        # FS_23_62_89_43: filter using L part of tangent token:
                        x.startswith(self.interp_ctx.parsed_ctx.tan_token_l_part)
                        # TODO: FS_06_99_43_60: Support list[str] - what if one type can have array of values (and we need to match any as in OR)?
                    )
                ]
                if proposed_values:
                    # Collect only until the first proposed value set from missing args:
                    break

        return proposed_values

    def _remaining_prop_values_from_given_prop_name(
        self,
        prop_name: str,
    ) -> list[str]:
        """
        # TODO: TODO_66_09_41_16: clarify command line processing
        #       Maybe we should elaborate logic and give different lists depending on `CompScope`.
        #       *   Current: If initial, use remaining.
        #       *   Extra: If subsequent, list all possible for given `prop_name`.
        """
        proposed_values: list[str] = []

        # Return filtered value set from the next missing arg:
        if prop_name in self.interp_ctx.curr_container.remaining_prop_name_to_prop_value:
            # TODO: TODO_66_09_41_16: clarify command line processing
            #       only one condition should be enough:
            #       `prop_name` is either in assigned or in remaining, not in both:
            assert prop_name not in self.interp_ctx.curr_container.assigned_prop_name_to_prop_value

            proposed_values = [
                # FS_71_87_33_52: `help_hint`:
                self.interp_ctx.help_hint_cache.get_value_with_help_hint(prop_name, x)
                for x in self.interp_ctx.curr_container.remaining_prop_name_to_prop_value[prop_name]
                if (
                    # TODO: TODO_66_09_41_16: clarify command line processing
                    #       This does not make sense - only str are expected.
                    #       Clean this up (or, at least, keep assert for a while).
                    # NOTE: This checks for `str` meaning primitive (scalar, not `list` or `dict`),
                    #       but this also filters out `int` or others.
                    #       Primitive types should be converted to `str` when loaded.
                    #       See `ConfigOnlyLoader.convert_envelope_fields_to_string` for example.
                    isinstance(x, str)
                    and
                    # FS_32_05_46_00: using `startswith`:
                    # FS_23_62_89_43: filter using L part of tangent token:
                    x.startswith(self.interp_ctx.parsed_ctx.tan_token_l_part)
                    # TODO: FS_06_99_43_60: Support list[str] - what if one type can have array of values (and we need to match any as in OR)?
                )
            ]

        return proposed_values



from __future__ import annotations

import os
from dataclasses import (
    dataclass,
    field,
)

import argrelay
from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_lib_root.enum_desc.CompScope import CompScope
from argrelay_lib_root.enum_desc.CompType import CompType
from argrelay_lib_root.enum_desc.ServerAction import ServerAction
from argrelay_lib_root.enum_desc.TermColor import TermColor
from argrelay_lib_root.misc_helper_common import (
    eprint,
    get_argrelay_dir,
)

UNKNOWN_COMP_KEY: str = str(0)
"""
Constant used when client is called via `CallConv.CliArgsConv`
and there is no `COMP_KEY` env var to obtain the value
(unlike during `CallConv.EnvVarsConv`).
"""

COMP_LINE_env_var: str = "COMP_LINE"
COMP_POINT_env_var: str = "COMP_POINT"
COMP_TYPE_env_var: str = "COMP_TYPE"
COMP_KEY_env_var: str = "COMP_KEY"


@dataclass(frozen = True)
class ShellContext:
    """
    Immutable input data from a shell
    """

    command_line: str = field()
    cursor_cpos: int = field()
    comp_type: CompType = field()
    comp_key: str = field()
    is_debug_enabled: bool = field()
    input_data: str = field()

    def __post_init__(self):
        assert 0 <= self.cursor_cpos <= len(self.command_line), (
            "Is this complex command line? See: https://github.com/argrelay/argrelay/issues/108"
        )

    @classmethod
    def from_env(
        cls,
        argv: list[str],
    ) -> ShellContext:
        """
        See Bash docs on these env var:
        https://www.gnu.org/software/bash/manual/html_node/Bash-Variables.html
        """
        if COMP_LINE_env_var in os.environ:
            # See `CallConv.EnvVarsConv`:
            command_line = os.environ[COMP_LINE_env_var]
            cursor_cpos = int(os.environ[COMP_POINT_env_var])
            # If `COMP_LINE_env_var` exists, the call is definitely NOT for `CompType.InvokeAction`:
            comp_type = CompType(int(os.environ.get(COMP_TYPE_env_var, CompType.SubsequentHelp.value)))
            assert comp_type != CompType.InvokeAction
            comp_key = os.environ.get(COMP_KEY_env_var, UNKNOWN_COMP_KEY)
        else:
            # See `CallConv.CliArgsConv`:
            argv = [os.path.basename(argv[0])] + argv[1:]
            command_line = " ".join(argv)
            cursor_cpos = len(command_line)
            # If `COMP_LINE_env_var` is missing, the call is definitely for `CompType.InvokeAction`
            comp_type = CompType.InvokeAction
            comp_key = UNKNOWN_COMP_KEY

        is_debug_enabled = (
            "ARGRELAY_DEBUG" in os.environ
            and
            "p" in os.environ["ARGRELAY_DEBUG"]
        )

        return cls(
            command_line = command_line,
            cursor_cpos = cursor_cpos,
            comp_type = comp_type,
            comp_key = comp_key,
            is_debug_enabled = is_debug_enabled,
            input_data = None,
        )

    def print_debug(
        self,
        end_str: str = "\n",
    ) -> None:
        if not self.is_debug_enabled:
            return
        eprint(TermColor.debug_output.value, end = "")
        eprint(f"\"{self.command_line}\"", end = " ")
        eprint(f"cursor_cpos: {self.cursor_cpos}", end = " ")
        eprint(f"comp_type: {self.comp_type}", end = " ")
        eprint(TermColor.reset_style.value, end = end_str)

    def create_call_context(
        self,
    ) -> CallContext:
        server_action: ServerAction = select_server_action(self.comp_type)
        return CallContext(
            client_version = argrelay.__version__,
            client_conf_target = get_client_conf_target(),
            server_action = server_action,
            command_line = self.command_line,
            cursor_cpos = self.cursor_cpos,
            comp_scope = CompScope.from_comp_type(self.comp_type),
            client_uid = get_user_name(),
            client_pid = os.getpid(),
            is_debug_enabled = self.is_debug_enabled,
            input_data = self.input_data,
        )


def select_server_action(
    comp_type,
) -> ServerAction:
    if comp_type is CompType.DescribeArgs:
        return ServerAction.DescribeLineArgs
    if comp_type is CompType.InvokeAction:
        return ServerAction.RelayLineArgs

    return ServerAction.ProposeArgValues


def get_user_name():
    try:
        return os.getlogin()
    except OSError:
        return ""


def get_client_conf_target():
    conf_path = f"{get_argrelay_dir()}/conf"
    if os.path.islink(conf_path):
        return os.readlink(conf_path)
    else:
        return ""



from os import getcwd
from typing import Union
from unittest import TestCase

from argrelay_lib_root.misc_helper_common import set_argrelay_dir
from argrelay_test_infra.test_infra import change_to_known_repo_path


class BaseTestClass(TestCase):
    """
    Root test class with defaults for all tests.
    """

    confusing_result_presence_msg = "remove expected result to avoid confusion"

    test_single_line_no: Union[int, None] = None
    """
    Set this `test_single_line_no` to limit test case having `line_no()` with that single line only.

    Setting it to `int` (instead of `None`) anywhere will also fail the test case in `tearDownClass()`.
    """

    @classmethod
    def setUpClass(cls):
        with change_to_known_repo_path("."):
            # Set CWD to the project root = `argrelay_dir` (see FS_29_54_67_86 dir structure):
            set_argrelay_dir(getcwd())
        TestCase.setUpClass()

    @classmethod
    def tearDownClass(cls):
        TestCase.tearDownClass()
        assert cls.test_single_line_no is None, "see docstring for `test_single_line_no`"

    def setUp(self):
        super().setUp()
        self.maxDiff = None

    def tearDown(self):
        super().tearDown()

    def skip_test_when_line_is_not_expected(
        self,
        actual_test_single_line_no: int,
        expected_test_single_line_no: int,
    ):
        """
        Use this function to skip running all tests (fail them)
        except the one specified in `expected_test_single_line_no`.
        """

        self.test_single_line_no = actual_test_single_line_no
        if expected_test_single_line_no is not None:
            if actual_test_single_line_no != expected_test_single_line_no:
                raise AssertionError("skip test as line no does not match")



from argrelay_app_client.relay_client.ClientAbstract import ClientAbstract
from argrelay_app_client.relay_client.ClientCommandFactoryRemote import ClientCommandFactoryRemote
from argrelay_lib_root.enum_desc.ProcRole import ProcRole
from argrelay_schema_config_client.runtime_data_client_app.ClientConfig import ClientConfig


class ClientRemote(ClientAbstract):
    """
    Client remote to server

    It talks to `CustomFlaskApp` which is remote server in this case.

    See also `ClientLocal` and `LocalServer`.
    """

    def __init__(
        self,
        client_config: ClientConfig,
        proc_role: ProcRole,
        w_pipe_end,
        is_optimized_completion: bool,
        server_index: int,
    ):
        super().__init__(
            client_config,
            ClientCommandFactoryRemote(
                client_config,
                proc_role,
                w_pipe_end,
                is_optimized_completion,
                server_index,
            ),
        )
        assert not client_config.use_local_requests



from __future__ import annotations

from argrelay_app_server.composite_forest.CompositeForestExtractor import extract_zero_arg_interp_tree
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_server_plugin_core.plugin_interp.FirstArgInterpFactoryConfigSchema import (
    ignored_func_ids_list_,
)
from argrelay_lib_server_plugin_core.plugin_interp.InterpTreeInterpFactory import (
    InterpTreeInterp,
    InterpTreeInterpFactory,
)
from argrelay_schema_config_server.runtime_data_server_app.ServerConfig import ServerConfig


class FirstArgInterpFactory(InterpTreeInterpFactory):
    """
    `FirstArgInterp` was re-implemented in terms of `InterpTreeInterp` (FS_01_89_09_24).

    Implements FS_42_76_93_51 first interp.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )
        # Unlike base `InterpTreeInterpFactory` which uses `CompositeInfoType.interp_tree` for `interp_selector_tree`,
        # `interp_selector_tree` for `FirstArgInterpFactory` is `CompositeInfoType.zero_arg_interp_tree`:
        assert self.interp_selector_tree == {}, "`InterpTreeInterpFactory` is supposed to extract nothing by `plugin_instance_id` for `FirstArgInterpFactory`."
        self.interp_selector_tree = extract_zero_arg_interp_tree(
            server_config.server_plugin_control.composite_forest,
        )

        self.mapped_func_ids: set[str] = set()
        """
        All `func_id`-s mapped somewhere within `composite_forest`.
        """

        self.index_props: set[str] = set()
        """
        All `index_prop`-s used by all loaded funcs.
        """

        self.func_data_envelopes: list[dict] = []
        """
        All func `data_envelopes` loaded.
        """

    def activate_plugin(
        self,
    ) -> None:
        """
        # NOTE: FS_42_76_93_51 first interp special case:
        # TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor
                TODO: Think of anther way because `FirstArgInterp` is obsolete - this role here only extends its agony.
                      `FirstArgInterp` is used as the root and single instance which
                      triggers func loading of itself (and all others recursively).

        Takes part in implementation of FS_01_89_09_24 interp tree.
        """

        # Build a map from `func_id` to `func_envelope`:
        func_ids_to_func_envelopes: dict[str, dict] = {}
        for delegator_plugin_id, action_delegator in self.server_config.action_delegators.items():
            func_envelopes = action_delegator.get_supported_func_envelopes()
            for func_envelope in func_envelopes:
                func_id = func_envelope[ReservedPropName.func_id.name]
                if func_id in func_ids_to_func_envelopes:
                    raise RuntimeError(
                        f"plugin_instance_id='{self.plugin_instance_id}': func_id='{func_id}' is published more than once by delegator_plugin_id='{delegator_plugin_id}'"
                    )
                func_ids_to_func_envelopes[func_id] = func_envelope

        self.load_interp_tree_abs_paths([])

        interp_tree_abs_path: tuple[str, ...] = tuple([])
        (
            mapped_func_ids,
            index_props,
            func_data_envelopes,
        ) = self.load_func_envelopes(
            interp_tree_abs_path,
            func_ids_to_func_envelopes,
        )

        self.mapped_func_ids.update(mapped_func_ids)
        self.index_props.update(index_props)
        self.func_data_envelopes.extend(func_data_envelopes)

        ignored_func_ids_list = self.plugin_config_dict.get(ignored_func_ids_list_, [])
        for func_id in func_ids_to_func_envelopes.keys():
            # TODO: TODO_19_67_22_89: remove `ignored_func_ids_list` - load as `FuncState.fs_unplugged`:
            if func_id not in mapped_func_ids:
                if func_id not in ignored_func_ids_list:
                    raise RuntimeError(
                        f"plugin_instance_id=`{self.plugin_instance_id}`: func_id=`{func_id}` is neither mapped anywhere nor is in `{ignored_func_ids_list_}`"
                    )
                    # TODO: Could it be this?
                    if False:
                        f"plugin_instance_id=`{self.plugin_instance_id}`: func_id=`{func_id}` is not published by any enabled delegator activated so far - please check `{plugin_enabled_}` and `{plugin_dependencies_}`"
                else:
                    # This `func_id` is ignored - skip:
                    continue
            else:
                if func_id in ignored_func_ids_list:
                    raise RuntimeError(
                        f"plugin_instance_id=`{self.plugin_instance_id}`: func_id=`{func_id}` is already mapped but still listed in `{ignored_func_ids_list_}`"
                    )
                else:
                    # This `func_id` is mapped - skip:
                    continue

    def is_root_func_loader(
        self,
    ) -> bool:
        return True

    def get_func_dynamic_index_props(
        self,
    ) -> list[str]:
        return list(self.index_props)

    def get_func_data_envelopes(
        self,
    ) -> list[dict]:
        return self.func_data_envelopes

    def load_interp_tree_abs_paths(
        self,
        this_plugin_instance_interp_tree_abs_paths: list[tuple[str, ...]],
    ):
        # TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor:
        #       At the moment, `FirstArgInterpFactory` is not plugged into any tree
        #       (while it should be part of composite forest likely as `CompositeNodeType.zero_arg_node`).
        #       Instead, it is selected via `first_interp_factory_id` in server config.
        #
        #       When this func is called for `FirstArgInterpFactory`, it cannot be found plugged anywhere
        #       within interp tree - instead, it walks the tree and invokes this func for other interps.
        super().load_interp_tree_abs_paths([])

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> FirstArgInterp:
        return FirstArgInterp(
            self.plugin_instance_id,
            # FS_42_76_93_51 first interp special case:
            # We use multiple config clones (with customization) `tree_node_config_dict` for interps at
            # different location on the FS_01_89_09_24 interp tree - these configs get cloned on `load_func_envelopes`.
            # But first interp is always at the root of the interp tree.
            # More over, first interp is the only one who triggers cascading calls to `load_func_envelopes`,
            # but never receives such call itself.
            # Instead of using `interp_tree_node_config_dict`, use `plugin_config_dict` directly:
            self.plugin_config_dict,
            interp_ctx,
            self.interp_selector_tree,
        )


class FirstArgInterp(InterpTreeInterp):
    """
    Dispatch command line interpretation to the next interpreter based on the command_id.

    `FirstArgInterp` was re-implemented in terms of `InterpTreeInterp` (FS_01_89_09_24).

    Implements FS_42_76_93_51 first interp.
    """

    def __init__(
        self,
        interp_factory_id: str,
        interp_tree_node_config_dict: dict,
        interp_ctx: InterpContext,
        interp_selector_tree: dict,
    ):
        super().__init__(
            interp_factory_id,
            interp_tree_node_config_dict,
            interp_ctx,
            interp_selector_tree,
        )



from enum import Enum


class DataOperation(Enum):
    """
    `DataOperation` defines server API to manipulate data at its backend.

    Implements FS_74_69_61_79 get set data envelope.

    For user operations on command line see `ServerAction`.
    """

    get_data_envelopes = "/get_data_envelopes/"

    set_data_envelopes = "/set_data_envelopes/"



from enum import IntEnum


class CompType(IntEnum):
    """
    These enum items are codes set by Bash into `COMP_TYPE` env var on different completion actions.

    In principle, there are 3 behaviors of the client (rather than behaviors of Bash invoking it):
    *   `PrefixShown` == `PrefixHidden` == `MenuCompletion`:
        provide all values suitable for the given command line and cursor position.
    *   `SubsequentHelp`:
        possibly provide more values.
    *   `DescribeArgs`:
        print how each arg is parsed from given command line and cursor position.

    These code given by Bash depends on (1) completion action by user and (2) Readline config.

    See Bash docs on `COMP_TYPE` env var:
    https://www.gnu.org/software/bash/manual/html_node/Bash-Variables.html

    This part of `~/.inputrc` configures Tab to do `complete` and Shift-Tab to do `menu-complete` completion actions:
    ```
    "\t": complete
    "\e[Z": menu-complete
    ```

    Other recommended settings for Readline:
    ```
    set show-all-if-ambiguous on
    set show-all-if-unmodified on
    set skip-completed-text on
    ```

    See Readline docs for details on `~/.inputrc` config:
    https://www.gnu.org/software/bash/manual/html_node/Readline-Init-File-Syntax.html
    """

    """
    This happens on first `complete` only when `set show-all-if-ambiguous on`.
    Subsequent `complete` results in `SubsequentHelp`.
    """
    PrefixShown = 33  # ASCII '!'

    """
    This happens on first `complete` only when `set show-all-if-ambiguous off`.
    In that case, it appears to work exactly like `PrefixShown`
    except Bash hides suggested values from user if these values are ambiguous
    (more than one value matching tangent `token_left_part`).
    """
    PrefixHidden = 9  # ASCII '\t'

    """
    This happens on subsequent `complete` actions (after the first `PrefixShown` or `PrefixHidden`).
    It is normally used as an opportunity for the client to suggest more values for completion.
    """
    SubsequentHelp = 63  # ASCII '?'

    """
    This happens on `menu-complete` - for example, on Shift-Tab given the `~/.inputrc` config above.
    Subsequent Shift-Tab-s do not invoke client - instead, Bash uses cashed suggestions and cycles through them.
    """
    # TODO: Test if it is feasible to cycle through options limited by current prefix (versus cycling through every item at current arg space).
    MenuCompletion = 37  # ASCII '%'

    """
    This `COMP_TYPE` value does not originate in Bash completion.
    Instead, it is set by `argrelay` Bash config script when invoked via special key combination (e.g. `ALT+SHIFT+Q`).
    It is used to describe args in the given command line.
    """
    DescribeArgs = 94  # ASCII '^'

    """
    This `COMP_TYPE` value does not originate in Bash completion.
    Instead, it is set by `argrelay` client for `CallConv.CliArgsConv`.
    """
    InvokeAction = 42  # ASCII '*'



from __future__ import annotations

import traceback
from typing import (
    cast,
    Union,
)

from argrelay_api_plugin_check_env_abstract.CheckEnvResult import CheckEnvResult
from argrelay_api_plugin_check_env_abstract.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_client.client_command_remote.ClientCommandRemoteWorkerJson import ClientCommandRemoteWorkerJson
from argrelay_app_client.client_pipeline.BytesHandlerJson import BytesHandlerJson
from argrelay_app_client.client_pipeline.BytesSrcLocal import BytesSrcLocal
from argrelay_app_client.client_spec.ShellContext import (
    ShellContext,
    UNKNOWN_COMP_KEY,
)
from argrelay_app_client.handler_response.ClientResponseHandlerCheckEnv import ClientResponseHandlerCheckEnv
from argrelay_app_client.relay_client.proc_worker import worker_main
from argrelay_lib_check_env_plugin_core.SchemaPluginCheckEvnServerResponseValueAbstract import (
    field_values_to_command_lines_,
    schema_plugin_check_evn_server_response_abstract_desc,
)
from argrelay_lib_root.enum_desc.CompType import CompType
from argrelay_lib_root.enum_desc.ProcRole import ProcRole
from argrelay_lib_root.enum_desc.ResultCategory import ResultCategory
from argrelay_lib_root.misc_helper_common import eprint
from argrelay_schema_config_client.runtime_data_client_app.ClientConfig import ClientConfig
from argrelay_schema_config_client.schema_config_client_app.ClientConfigSchema import client_config_desc


class PluginCheckEnvServerResponseValueAbstract(PluginCheckEnvAbstract):

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        self.response_value_config_desc = schema_plugin_check_evn_server_response_abstract_desc
        super().__init__(
            plugin_instance_id,
            plugin_config_dict,
        )

    def validate_config(
        self,
    ) -> None:
        self.response_value_config_desc.validate_dict(self.plugin_config_dict)

    def execute_check(
        self,
        online_mode: Union[bool, None],
    ) -> list[CheckEnvResult]:

        check_env_results: list[CheckEnvResult] = []
        for field_name, command_line in self.plugin_config_dict[field_values_to_command_lines_].items():
            client_config: ClientConfig = client_config_desc.obj_from_default_file()

            for server_index in range(len(client_config.redundant_servers)):

                indexed_field_name = f"{field_name}[{server_index}]"

                # noinspection PyBroadException
                try:
                    shell_ctx = ShellContext(
                        command_line = command_line,
                        cursor_cpos = len(command_line),
                        comp_type = CompType.InvokeAction,
                        comp_key = UNKNOWN_COMP_KEY,
                        is_debug_enabled = False,
                        input_data = None,
                    )

                    call_ctx = shell_ctx.create_call_context()

                    proc_role: ProcRole = ProcRole.CheckEnvWorker

                    try:
                        command_obj: ClientCommandRemoteWorkerJson = worker_main(
                            call_ctx,
                            client_config,
                            proc_role,
                            False,
                            None,
                            server_index,
                        )
                        is_offline = False
                    except ConnectionError as e:
                        # noinspection PySimplifyBooleanCheck
                        if online_mode == True:
                            raise e
                        else:
                            is_offline = True

                    if is_offline:
                        check_env_results.append(self.verify_offline(
                            indexed_field_name,
                        ))
                    else:
                        invocation_input: InvocationInput = cast(
                            ClientResponseHandlerCheckEnv,
                            cast(
                                BytesHandlerJson,
                                cast(
                                    BytesSrcLocal,
                                    command_obj.bytes_src,
                                ).bytes_handler,
                            ).client_response_handler,
                        ).invocation_input

                        check_env_results.append(self.verify_online_value(
                            indexed_field_name,
                            invocation_input.custom_plugin_data[field_name],
                        ))

                except Exception as e:
                    eprint(traceback.format_exc())
                    check_env_results.append(CheckEnvResult(
                        result_category = ResultCategory.ExecutionFailure,
                        result_key = indexed_field_name,
                        result_value = None,
                        result_message = f"Failed: {str(type(e))}: {str(e)}",
                    ))
        return check_env_results

    def verify_online_value(
        self,
        field_name,
        field_value,
    ) -> CheckEnvResult:
        """
        Default implementation verifies nothing.
        """
        return CheckEnvResult(
            result_category = ResultCategory.VerificationSuccess,
            result_key = field_name,
            result_value = field_value,
            result_message = None,
        )

    # noinspection PyMethodMayBeStatic
    def verify_offline(
        self,
        field_name,
    ) -> CheckEnvResult:
        """
        Default implementation verifies nothing.
        """
        return CheckEnvResult(
            result_category = ResultCategory.ServerOffline,
            result_key = field_name,
            result_value = None,
            result_message = "The server is offline",
        )



from __future__ import annotations

import time

from argrelay_lib_root.misc_helper_common import eprint


class ElapsedTime:
    """
    *   Instance holds a named timestamp.
    *   Class maintains a list of all instances (named timestamps) stored in order of measurements.

    At the moment, a static singleton is enough as perf measurements are done with single client.
    """

    all_ts: list[ElapsedTime] = []

    is_debug_enabled: bool = False
    """
    This is normally set based on `ShellContext.is_debug_enabled`.
    """

    def __init__(
        self,
        entry_name: str,
        entry_time: int,
    ):
        self.entry_name = entry_name
        self.entry_time = entry_time

    def print(self, prev_ts: int):
        ElapsedTime.print_formatted(
            self.entry_name,
            self.entry_time,
            self.entry_time - prev_ts,
        )

    @classmethod
    def clear_measurements(cls):
        cls.all_ts.clear()

    @staticmethod
    def print_formatted(name: str, ts: int, diff: int):
        eprint(f"{diff / 1_000_000_000:f}s: {name}")

    @classmethod
    def measure(cls, entry_name: str):
        cls.all_ts.append(ElapsedTime(entry_name, time.time_ns()))

    @classmethod
    def print_all(cls):
        eprint()
        if cls.all_ts:
            first_ts = cls.all_ts[0]
            prev_ts = first_ts
            prev_ts.print(prev_ts.entry_time)
            remaining = cls.all_ts[1:]
        else:
            first_ts = None
            prev_ts = None
            remaining = []

        for ts in remaining:
            ts.print(prev_ts.entry_time)
            prev_ts = ts

        # total:
        if cls.all_ts:
            prev_ts.print_formatted(
                "total",
                first_ts.entry_time,
                prev_ts.entry_time - first_ts.entry_time,
            )

    @classmethod
    def print_all_if_debug(cls):
        if cls.is_debug_enabled:
            cls.print_all()



from __future__ import annotations

import os
import time
from datetime import datetime
from typing import Union

from argrelay_api_plugin_check_env_abstract.CheckEnvResult import CheckEnvResult
from argrelay_api_plugin_check_env_abstract.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay_lib_check_env_plugin_core.check_env_utils import format_time_to_relative
from argrelay_lib_root.enum_desc.ResultCategory import ResultCategory


class PluginCheckEnvDevShell(PluginCheckEnvAbstract):
    """
    This plugin reports value of `ARGRELAY_DEV_SHELL` env var.
    """

    env_var_name: str = "ARGRELAY_DEV_SHELL"

    # noinspection PyMethodMayBeStatic
    def execute_check(
        self,
        online_mode: Union[bool, None],
    ) -> list[CheckEnvResult]:
        env_var_value = os.environ.get(self.env_var_name)

        if env_var_value is None:
            result_message = "The env var is not set => outside `@/exe/dev_shell.bash`."
        else:
            curr_time_ms = time.time() * 1000
            prev_time_ms = convert_iso_str_date_to_ms(env_var_value)
            if prev_time_ms is None:
                result_message = f"The env var is set => inside `@/exe/dev_shell.bash`."
            else:
                rel_time: str = format_time_to_relative(
                    curr_time_ms,
                    prev_time_ms,
                )
                result_message = f"~ {rel_time}: the env var is set => inside `@/exe/dev_shell.bash`."

        return [CheckEnvResult(
            result_category = ResultCategory.VerificationSuccess,
            result_key = self.env_var_name,
            result_value = str(env_var_value),
            result_message = result_message,
        )]


def convert_iso_str_date_to_ms(
    date_str: str,
) -> Union[int, None]:
    try:
        return int(datetime.fromisoformat(date_str).timestamp()) * 1000
    except:
        return None



from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_app_client.relay_client import ClientCommandAbstract
from argrelay_lib_root.enum_desc.ProcRole import ProcRole
from argrelay_lib_root.enum_desc.ServerAction import ServerAction


class ClientCommandFactoryAbstract:

    def create_command(
        self,
        call_ctx: CallContext,
    ) -> ClientCommandAbstract:
        pass


def select_client_response_handler(
    proc_role: ProcRole,
    server_action: ServerAction,
):
    if server_action is ServerAction.ProposeArgValues:
        """
        This handler is unused most of the time in favor of direct (optimized):
        `BytesHandlerTextProposeArgValuesOptimized`

        See `completion_perf_notes.md` for details.

        This non-optimized implementation is still useful to get access to
        internal server state in tests via `ClientLocal` - see FS_66_17_43_42 test_infra / special test mode #1.

        To enable use of this command, see:
        *   ClientConfig.optimize_completion_request
        *   LocalClientEnvMockBuilder
        """
        from argrelay_app_client.handler_response.ClientResponseHandlerProposeArgValues import (
            ClientResponseHandlerProposeArgValues
        )
        return ClientResponseHandlerProposeArgValues()
    if server_action is ServerAction.DescribeLineArgs:
        from argrelay_app_client.handler_response.ClientResponseHandlerDescribeLineArgs import (
            ClientResponseHandlerDescribeLineArgs
        )
        return ClientResponseHandlerDescribeLineArgs()
    if server_action is ServerAction.RelayLineArgs:
        if proc_role == ProcRole.CheckEnvWorker:
            from argrelay_app_client.handler_response.ClientResponseHandlerCheckEnv import (
                ClientResponseHandlerCheckEnv
            )
            return ClientResponseHandlerCheckEnv()
        else:
            from argrelay_app_client.handler_response.ClientResponseHandlerRelayLineArgs import (
                ClientResponseHandlerRelayLineArgs
            )
            return ClientResponseHandlerRelayLineArgs()
    raise RuntimeError



from __future__ import annotations

from marshmallow import (
    fields,
    RAISE,
    Schema,
)

from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_lib_server_plugin_demo.demo_git.GitRepoPropName import GitRepoPropName

repo_rel_path_ = "repo_rel_path"
is_repo_enabled_ = "is_repo_enabled"

load_repo_entries_ = "load_repo_entries"
load_repo_tags_ = "load_repo_tags"
load_repo_commits_ = "load_repo_commits"

load_tags_last_days_ = "load_tags_last_days"
load_commits_max_count_ = "load_commits_max_count"
envelope_properties_ = "envelope_properties"


class GitRepoEntryConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    repo_rel_path = fields.String(
        required = True,
    )

    load_repo_entries = fields.Boolean(
        required = False,
        load_default = None,
        allow_none = True,
    )

    load_repo_tags = fields.Boolean(
        required = False,
        load_default = None,
        allow_none = True,
    )

    load_repo_commits = fields.Boolean(
        required = False,
        load_default = None,
        allow_none = True,
    )

    load_tags_last_days = fields.Integer(
        required = False,
        load_default = 300,
    )

    load_commits_max_count = fields.Integer(
        required = False,
        load_default = 100,
    )

    is_repo_enabled = fields.Boolean(
        required = False,
        load_default = True,
    )

    # These `prop_name` and `prop_values` are copied into `data_envelope`:
    envelope_properties = fields.Dict(
        keys = fields.String(),
        values = fields.String(),
        required = True,
    )


git_repo_entry_config_desc = TypeDesc(
    dict_schema = GitRepoEntryConfigSchema(),
    ref_name = GitRepoEntryConfigSchema.__name__,
    dict_example = {
        repo_rel_path_: "argrelay.git",
        is_repo_enabled_: True,
        load_repo_tags_: False,
        load_repo_commits_: False,
        load_tags_last_days_: 1,
        load_commits_max_count_: 10,
        envelope_properties_: {
            GitRepoPropName.git_repo_alias.name: "ar",
        }
    },
    default_file_path = "",
)



from __future__ import annotations

from typing import (
    Type,
    Union,
)

from argrelay_api_plugin_server_abstract.DelegatorAbstract import DelegatorAbstract
from argrelay_api_server_cli.schema_response.ArgValuesSchema import arg_values_
from argrelay_api_server_cli.schema_response.AssignedValue import AssignedValue
from argrelay_app_client.client_command_remote.ClientCommandRemoteWorkerJson import ClientCommandRemoteWorkerJson
from argrelay_app_client.relay_client import __main__
from argrelay_lib_root.enum_desc.CompType import CompType
from argrelay_test_infra.test_infra import parse_line_and_cpos
from argrelay_test_infra.test_infra.ClientServerTestClass import ClientServerTestClass
from argrelay_test_infra.test_infra.EnvMockBuilder import EnvMockBuilder


class RemoteTestClass(ClientServerTestClass):
    """
    Supports FS_66_17_43_42 test_infra / special test mode #2.

    This class is similar to `LocalTestClass` as it is supposed to use `EnvMockBuilder`
    to capture server response, but, unlike `LocalTestClass`, requests go to separate OS process with running server.
    """

    # TODO: Allow intercepting subprocess.* invocation and asserting their input (e.g. whether specific command was invoked).
    def verify_output_via_remote_client(
        self,
        test_line: str,
        comp_type: CompType,
        expected_suggestions: Union[list[str], None],
        container_ipos_to_expected_assignments: Union[dict[int, dict[str, AssignedValue]], None],
        delegator_class: Union[Type[DelegatorAbstract], None],
        envelope_ipos_to_prop_values: Union[dict[int, dict[str, str]], None],
        expected_container_ipos_to_used_token_bucket: Union[dict[int, Union[int, None]], None],
        init_env_mock_builder: EnvMockBuilder,
    ):
        (command_line, cursor_cpos) = parse_line_and_cpos(test_line)

        if envelope_ipos_to_prop_values is not None:
            self.assertIsNotNone(delegator_class)

        # TODO: Make it possible to capture response for other `ServerAction`-s.
        #       At the moment, only `ServerAction.RelayLineArgs` is supported
        #       (because we are intercepting `invoke_action` on target `delegator_class`).
        #       But we could intercept on `*ClientResponseHandler.handle_response`:
        #       if non-optimized command is used for `ServerAction.ProposeArgValues`,
        #       it will always be JSON for all operations.
        assert comp_type is CompType.InvokeAction
        assert delegator_class is not None

        if delegator_class is not None:
            init_env_mock_builder.set_capture_delegator_invocation_input(delegator_class)

        env_mock_builder = (
            init_env_mock_builder
            .set_command_line(command_line)
            .set_cursor_cpos(cursor_cpos)
            .set_comp_type(comp_type)
        )
        with env_mock_builder.build():
            __main__.main()

            command_obj = __main__.main()
            assert isinstance(command_obj, ClientCommandRemoteWorkerJson)

            call_ctx = command_obj.call_ctx

            # TODO: TODO_32_99_70_35: (JSONPath?) Currently, this verifier ensures what things exists.
            #       Add a way to ensure what things do not exist.
            #       For example,
            #       * no `data_envelope` number N.
            #       * `data_envelope` does not have field A.
            #       * `data_envelope` does not have value 'x' for field A.

            actual_suggestions = None
            if expected_suggestions is not None:
                # TODO: This if-branch is unused - see if it is useful or clean up:
                # TODO: Consider intercepting `*ClientResponseHandler.handle_response`.
                #       At the moment, only intercepts for `ServerAction.RelayLineArgs` are supported
                #       (via `invoke_action` on target `delegator_class`) which make this access to
                #       response for `ServerAction.ProposeArgValues` fail:
                actual_suggestions = command_obj.response_dict[arg_values_]

            envelope_containers = None
            if delegator_class:
                envelope_containers = EnvMockBuilder.invocation_input.envelope_containers

            self.verify_response_data(
                call_ctx,
                actual_suggestions,
                container_ipos_to_expected_assignments,
                None,
                delegator_class,
                envelope_ipos_to_prop_values,
                expected_suggestions,
                envelope_containers,
                expected_container_ipos_to_used_token_bucket,
            )



"""
This module lists interfaces for various args.
"""

from __future__ import annotations


class AbstractArg:
    """
    An arg of any origin (env, config, command) which can be used for a func.
    """


class ArgCommand(AbstractArg):
    """
    Represents `command_arg` - see FS_27_16_67_19 line syntax
    """

    def get_arg_tokens(
        self,
    ) -> list[int]:
        return []


class ArgCommandValue(ArgCommand):
    """
    Represents any `command_arg` with `arg_value`.
    """

    def get_arg_value(
        self
    ) -> str:
        raise NotImplementedError


class ArgCommandName(ArgCommand):
    """
    Represents any `command_arg` with `arg_name`.
    """

    def get_arg_name(
        self
    ) -> str:
        raise NotImplementedError


class ArgCommandValueOffered(ArgCommandValue):
    """
    Represents FS_96_46_42_30 `offered_arg`.
    """

    def get_arg_token(
        self,
    ) -> int:
        assert len(self.get_arg_tokens()) == 1
        return self.get_arg_tokens()[0]


class ArgCommandValueDictated(ArgCommandName, ArgCommandValue):
    """
    Represents FS_20_88_05_60 `dictated_arg`.
    """


class ArgCommandIncomplete(ArgCommandName):
    """
    Represents FS_08_58_30_24 `incomplete_arg`.
    """



from __future__ import annotations

import dataclasses
import sys

from argrelay_api_plugin_server_abstract.delegator_utils import redirect_to_not_disambiguated_error
from argrelay_api_plugin_server_abstract.DelegatorAbstract import (
    get_func_id_from_interp_ctx,
    get_func_id_from_invocation_input,
)
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_client.relay_client.__main__ import run_client
from argrelay_app_server.mongo_data.ProgressTracker import ProgressTracker
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.relay_server.QueryEngine import populate_query_dict
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_root.enum_desc.SpecialFunc import SpecialFunc
from argrelay_lib_root.misc_helper_common import eprint
from argrelay_lib_server_plugin_core.plugin_delegator.client_invocation_utils import prohibit_unconsumed_args
from argrelay_lib_server_plugin_core.plugin_delegator.DelegatorDataBackendBase import (
    DelegatorDataBackendBase,
    get_collection_search_control,
)
from argrelay_schema_config_client.runtime_data_client_app.ClientConfig import ClientConfig
from argrelay_schema_config_client.schema_config_client_app.ClientConfigSchema import client_config_desc
from argrelay_schema_config_server.runtime_data_server_app.EnvelopeCollection import EnvelopeCollection
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    data_envelope_desc,
    instance_data_,
)
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    func_id_,
    search_control_list_,
)

collection_name_container_ipos_ = 1
data_envelope_container_ipos_ = 2


class DelegatorDataBackendSet(DelegatorDataBackendBase):
    """
    Implements `SpecialFunc.func_id_set_data_envelopes` for FS_74_69_61_79 get set data envelope.
    """

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:

        collection_search_control = get_collection_search_control(
        )

        func_envelopes = []

        given_function_envelope = {
            instance_data_: {
                func_id_: SpecialFunc.func_id_set_data_envelopes.name,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                    collection_search_control,
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
            ReservedPropName.help_hint.name: "Set `data_envelope`-s based on their `index_prop`-s.",
            ReservedPropName.func_state.name: FuncState.fs_alpha.name,
            ReservedPropName.func_id.name: SpecialFunc.func_id_set_data_envelopes.name,
        }
        func_envelopes.append(given_function_envelope)

        return func_envelopes

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:

        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        func_id = get_func_id_from_interp_ctx(interp_ctx)
        assert func_id == SpecialFunc.func_id_set_data_envelopes.name

        vararg_container_ipos = data_envelope_container_ipos_

        # Verify that func is selected:
        if interp_ctx.curr_container_ipos < vararg_container_ipos:
            return redirect_to_not_disambiguated_error(
                interp_ctx,
                local_server.plugin_config,
                ReservedEnvelopeClass.class_collection.name,
            )

        collection_container = interp_ctx.envelope_containers[collection_name_container_ipos_]
        vararg_container = interp_ctx.envelope_containers[vararg_container_ipos]

        # FS_74_69_61_79 get set data envelope:
        # The way `get` is implemented via double `ServerAction.RelayLineArgs` call:
        # *   The first request to the server without `input_data` sends invocation data to the client.
        # *   The client runs the second request with `input_data` collected (e.g. from `stdin`).

        if interp_ctx.parsed_ctx.input_data is None:
            # 1st request
            pass

        else:
            # 2nd request

            # All what is left to do is to delete and store objects.

            # Delete:
            collection_name = collection_container.assigned_prop_name_to_prop_value[
                ReservedPropName.collection_name.name
            ].prop_value
            query_dict = populate_query_dict(vararg_container)
            eprint(f"delete: {query_dict}")
            local_server.delete_data_envelopes(
                collection_name,
                query_dict,
            )

            # Store:
            data_envelopes = []
            for json_line in interp_ctx.parsed_ctx.input_data.splitlines():
                if len(json_line.strip()) == 0:
                    continue
                data_envelopes.append(data_envelope_desc.obj_from_yaml_str(json_line))
            envelope_collection = EnvelopeCollection(
                collection_name = collection_name,
                data_envelopes = data_envelopes,
            )
            eprint(f"envelope_collection: {envelope_collection}")
            progress_tracker = ProgressTracker()
            local_server.store_mongo_data_step(
                [envelope_collection],
                "",
                progress_tracker,
            )

            # Post-validate:
            local_server.post_validate_collection_for_missing_index_prop_names(
                collection_name,
                progress_tracker,
            )

            # Invalidate cache:
            local_server.invalidate_cache_for_collection(collection_name)

        # Plugin to invoke on client side:
        delegator_plugin_instance_id = self.plugin_instance_id
        # Package into `InvocationInput` payload object:
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = local_server.plugin_config.server_plugin_instances[
                delegator_plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        func_id = get_func_id_from_invocation_input(invocation_input)
        assert func_id == SpecialFunc.func_id_set_data_envelopes.name

        prohibit_unconsumed_args(invocation_input)

        if invocation_input.call_ctx.input_data is None:
            # 1st response

            call_ctx = dataclasses.replace(
                invocation_input.call_ctx,
                # FS_74_69_61_79 get set data envelope:
                # Load data from stdin and call REST API to update data on the server:
                input_data = sys.stdin.read(),
            )

            if sys.stdin.isatty():
                eprint("provide `data_envelope` one JSON per line:")

            client_config: ClientConfig = client_config_desc.obj_from_default_file()
            if len(client_config.redundant_servers) > 1:
                eprint("WARN: this command may only update data for one of the `redundant_servers` in `client_config`")

            run_client(
                client_config,
                call_ctx,
            )
        else:
            # 2nd response
            pass



from __future__ import annotations

from argrelay_app_server.composite_forest.CompositeForestWalker import (
    CompositeForestWalkerAbstract,
    TraverseDecision,
)
from argrelay_lib_root.enum_desc.PluginType import PluginType
from argrelay_lib_server_plugin_core.plugin_interp.FuncTreeInterpFactory import FuncTreeInterpFactory
from argrelay_lib_server_plugin_core.plugin_interp.InterpTreeInterpFactory import InterpTreeInterpFactory
from argrelay_schema_config_server.runtime_data_server_app.CompositeForest import CompositeForest
from argrelay_schema_config_server.runtime_data_server_app.CompositeNode import (
    BaseNode,
    FuncTreeNode,
    InterpTreeNode,
    TreePathNode,
    ZeroArgNode,
)
from argrelay_schema_config_server.runtime_data_server_app.CompositeNodeType import CompositeNodeType


def validate_composite_forest(
    composite_forest: CompositeForest,
    plugin_instances: dict[str, "AbstractPluginServer"],
):
    """
    Execute all validations for the composite forest.
    """

    for forest_validator_class in [
        _CompositeForestValidator_zero_arg_node_is_root,
        _CompositeForestValidator_func_tree_node_is_leaf,
        _CompositeForestValidator_zero_arg_node_with_interp_tree_node,
        _CompositeForestValidator_interp_tree_node_with_func_tree_node,
        _CompositeForestValidator_plugin_instance_id,
    ]:
        forest_validator_instance = forest_validator_class(
            composite_forest,
            plugin_instances,
        )
        forest_validator_instance.walk_tree_roots()


class CompositeForestValidatorAbstract(CompositeForestWalkerAbstract):
    """
    Base class to validate FS_33_76_82_84 composite forest.
    """

    def __init__(
        self,
        composite_forest: CompositeForest,
        plugin_instances: dict[str, "AbstractPluginServer"],
    ):
        # main input:
        super().__init__(
            composite_forest,
        )
        self.plugin_instances = plugin_instances


# noinspection PyPep8Naming
class _CompositeForestValidator_zero_arg_node_is_root(CompositeForestValidatorAbstract):
    """
    Validate that:
    *   all `CompositeNodeType.zero_arg_node` are only at tree roots within the composite forest
    *   all tree roots are `CompositeNodeType.zero_arg_node`
    """

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        self._assert_is_root()

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        self._assert_is_not_root()

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        self._assert_is_not_root()

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        self._assert_is_not_root()

    def _assert_is_root(self):
        # TODO: replace silent asserts with message describing location of the problem:
        assert len(self.curr_path) == 1
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _assert_is_not_root(self):
        assert len(self.curr_path) != 1
        self.visitor_decision = TraverseDecision.walk_sub_tree


# noinspection PyPep8Naming
class _CompositeForestValidator_func_tree_node_is_leaf(CompositeForestValidatorAbstract):
    """
    Validate that:
    *   all `CompositeNodeType.func_tree_node` are leaves only within the composite forest
    *   all tree leaves are `CompositeNodeType.func_tree_node`
    """

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        self._assert_is_not_leaf()

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        self._assert_is_not_leaf()

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        self._assert_is_not_leaf()

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        self._assert_is_leaf()

    def _assert_is_leaf(self):
        assert self.curr_node.sub_tree is None
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _assert_is_not_leaf(self):
        assert self.curr_node.sub_tree is not None
        self.visitor_decision = TraverseDecision.walk_sub_tree


# noinspection PyPep8Naming
class _CompositeForestValidator_zero_arg_node_with_interp_tree_node(CompositeForestValidatorAbstract):
    """
    Validate that:
    *   Descendants of `CompositeNodeType.zero_arg_node` are `CompositeNodeType.interp_tree_node`
        (separated by 0 or N depths of only `CompositeNodeType.tree_path_node`-s)
    *   Ancestor of `CompositeNodeType.interp_tree_node` is `CompositeNodeType.zero_arg_node`
        (separated by 0 or N depths of only `CompositeNodeType.tree_path_node`-s)
    """

    def walk_tree_roots(
        self,
    ) -> None:
        # noinspection PyAttributeOutsideInit
        self.non_tree_path_ancestor_nodes_stack: [BaseNode] = []
        super().walk_tree_roots()

    ####################################################################################################################
    # zero_arg_node

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        self._push_non_tree_path_node_ancestor()
        self._assert_children_are_interp_tree_node_or_tree_path_node()

    def _leave_zero_arg_node(self, node: ZeroArgNode) -> None:
        self._pop_non_tree_path_node_ancestor()

    ####################################################################################################################
    # tree_path_node

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        if self.non_tree_path_ancestor_nodes_stack:
            if self.non_tree_path_ancestor_nodes_stack[-1].node_type is CompositeNodeType.zero_arg_node:
                self._assert_children_are_interp_tree_node_or_tree_path_node()

    ####################################################################################################################
    # interp_tree_node

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        self._assert_non_tree_path_node_ancestor_is_zero_arg_node()
        self._push_non_tree_path_node_ancestor()

    def _leave_interp_tree_node(self, node: InterpTreeNode) -> None:
        self._pop_non_tree_path_node_ancestor()

    ####################################################################################################################
    # func_tree_node

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        self._push_non_tree_path_node_ancestor()

    def _leave_func_tree_node(self, node: FuncTreeNode) -> None:
        self._pop_non_tree_path_node_ancestor()

    ####################################################################################################################

    def _assert_is_leaf(self):
        assert self.curr_node.sub_tree is None
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _assert_children_are_interp_tree_node_or_tree_path_node(self):
        assert self.curr_node.sub_tree is not None
        for child_node in self.curr_node.sub_tree.values():
            assert (
                child_node.node_type is CompositeNodeType.interp_tree_node
                or
                child_node.node_type is CompositeNodeType.tree_path_node
            )
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _assert_non_tree_path_node_ancestor_is_zero_arg_node(self):
        assert self.non_tree_path_ancestor_nodes_stack[-1].node_type is CompositeNodeType.zero_arg_node
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _push_non_tree_path_node_ancestor(self):
        assert self.curr_node.node_type is not CompositeNodeType.tree_path_node
        self.non_tree_path_ancestor_nodes_stack.append(self.curr_node)
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _pop_non_tree_path_node_ancestor(self):
        self.non_tree_path_ancestor_nodes_stack.pop()
        self.visitor_decision = TraverseDecision.walk_sub_tree


# noinspection PyPep8Naming
class _CompositeForestValidator_interp_tree_node_with_func_tree_node(CompositeForestValidatorAbstract):
    """
    Validate that:
    *   Descendants of `CompositeNodeType.interp_tree_node` are `CompositeNodeType.func_tree_node`
        (separated by 0 or N depths of only `CompositeNodeType.tree_path_node`-s)
    *   Ancestor of `CompositeNodeType.func_tree_node` is `CompositeNodeType.interp_tree_node`
        (separated by 0 or N depths of only `CompositeNodeType.tree_path_node`-s)
    """

    def walk_tree_roots(
        self,
    ) -> None:
        # noinspection PyAttributeOutsideInit
        self.non_tree_path_ancestor_nodes_stack: [BaseNode] = []
        super().walk_tree_roots()

    ####################################################################################################################
    # zero_arg_node

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        self._push_non_tree_path_node_ancestor()

    def _leave_zero_arg_node(self, node: ZeroArgNode) -> None:
        self._pop_non_tree_path_node_ancestor()

    ####################################################################################################################
    # tree_path_node

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        if self.non_tree_path_ancestor_nodes_stack:
            if self.non_tree_path_ancestor_nodes_stack[-1].node_type is CompositeNodeType.interp_tree_node:
                self._assert_children_are_func_tree_node_or_tree_path_node()

    ####################################################################################################################
    # interp_tree_node

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        self._push_non_tree_path_node_ancestor()
        self._assert_children_are_func_tree_node_or_tree_path_node()

    def _leave_interp_tree_node(self, node: InterpTreeNode) -> None:
        self._pop_non_tree_path_node_ancestor()

    ####################################################################################################################
    # func_tree_node

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        self._assert_non_tree_path_node_ancestor_is_interp_tree_node()
        self._push_non_tree_path_node_ancestor()

    def _leave_func_tree_node(self, node: FuncTreeNode) -> None:
        self._pop_non_tree_path_node_ancestor()

    ####################################################################################################################

    def _assert_is_leaf(self):
        assert self.curr_node.sub_tree is None
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _assert_children_are_func_tree_node_or_tree_path_node(self):
        assert self.curr_node.sub_tree is not None
        for child_node in self.curr_node.sub_tree.values():
            assert (
                child_node.node_type is CompositeNodeType.func_tree_node
                or
                child_node.node_type is CompositeNodeType.tree_path_node
            )
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _assert_non_tree_path_node_ancestor_is_interp_tree_node(self):
        assert self.non_tree_path_ancestor_nodes_stack[-1].node_type is CompositeNodeType.interp_tree_node
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _push_non_tree_path_node_ancestor(self):
        assert self.curr_node.node_type is not CompositeNodeType.tree_path_node
        self.non_tree_path_ancestor_nodes_stack.append(self.curr_node)
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _pop_non_tree_path_node_ancestor(self):
        self.non_tree_path_ancestor_nodes_stack.pop()
        self.visitor_decision = TraverseDecision.walk_sub_tree


# noinspection PyPep8Naming
class _CompositeForestValidator_plugin_instance_id(CompositeForestValidatorAbstract):
    """
    Validate that:
    *   all `CompositeNodeType.zero_arg_node`-s use `PluginType.InterpFactoryPlugin`
    *   all `CompositeNodeType.interp_tree_node`-s use `PluginType.InterpFactoryPlugin`
    """

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        self._assert_interp_factory_plugin(InterpTreeInterpFactory)

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        self._assert_nothing()

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        self._assert_interp_factory_plugin(FuncTreeInterpFactory)

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        self._assert_nothing()

    def _assert_nothing(self):
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _assert_interp_factory_plugin(self, base_class):
        assert self.curr_node.plugin_instance_id is not None

        plugin_instance = self.plugin_instances[self.curr_node.plugin_instance_id]

        assert plugin_instance.get_plugin_type() is PluginType.InterpFactoryPlugin

        # This validation may be relaxed, but now it is how working composite forests are built:
        assert issubclass(type(plugin_instance), base_class)

        self.visitor_decision = TraverseDecision.walk_sub_tree



import signal
from dataclasses import asdict

from argrelay_api_server_cli.schema_request.CallContextSchema import call_context_desc
from argrelay_api_server_cli.server_spec.const_int import BASE_URL_FORMAT
from argrelay_app_client.client_command_remote.ClientCommandRemoteWorkerAbstract import (
    ClientCommandRemoteWorkerAbstract,
)
from argrelay_lib_root.misc_helper_common import eprint
from argrelay_lib_root.misc_helper_common.ElapsedTime import ElapsedTime

has_error_happened = False


def _signal_handler(signal_number, signal_frame):
    if signal_number == signal.SIGALRM:
        # Import hanged:
        eprint("ERROR: `import` hanged, `Ctrl + C` and retry, details: https://github.com/argrelay/argrelay/issues/89")
        global has_error_happened
        has_error_happened = True


class ClientCommandRemoteWorkerJson(ClientCommandRemoteWorkerAbstract):

    def _execute_remote_call(
        self,
    ):
        server_url = BASE_URL_FORMAT.format(
            **asdict(self.curr_connection_config)
        ) + f"{self.call_ctx.server_action.value}"
        headers_dict = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        # NOTE: So far, single `CallContextSchema` is reused for all client requests:
        request_json = call_context_desc.dict_schema.dumps(self.call_ctx)
        ElapsedTime.measure("before_request")

        # TODO: TODO_30_69_19_14: infinite spinner:
        #     There is a bug - if a child is used, it deadlocks occasionally while importing `requests`:
        #     https://github.com/argrelay/argrelay/issues/89
        #     The workaround is to abort (Ctrl+C) and retry - but this is annoying.
        # Attempt to detect hanging import by setting an alarm at least (not resolving it yet):
        signal.signal(signal.SIGALRM, _signal_handler)
        signal.alarm(1)
        import requests
        signal.alarm(0)

        try:
            response_obj = requests.post(
                server_url,
                headers = headers_dict,
                data = request_json,
            )
        except requests.exceptions.ConnectionError as e:
            # translate to builtin:
            raise ConnectionError(e)

        ElapsedTime.measure("after_request")
        try:
            if response_obj.ok:
                self.bytes_src.accept_bytes(response_obj.content)
            else:
                self.raise_error(response_obj.status_code)
        finally:
            ElapsedTime.measure("after_handle_response")



"""
Python part of FS_36_17_84_44 `check_env` implementation.
"""

import os
import sys

from argrelay_api_plugin_check_env_abstract.CheckEnvResult import CheckEnvResult
from argrelay_api_plugin_check_env_abstract.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay_api_plugin_client_abstract.PluginClientAbstract import instantiate_client_plugin
from argrelay_app_client.relay_client.client_utils import handle_main_exception
from argrelay_lib_root import misc_helper_common
from argrelay_lib_root.enum_desc.OutputCategory import OutputCategory
from argrelay_lib_root.enum_desc.PluginSide import PluginSide
from argrelay_lib_root.enum_desc.PluginType import PluginType
from argrelay_lib_root.enum_desc.ResultCategory import ResultCategory
from argrelay_lib_root.enum_desc.TermColor import TermColor
from argrelay_schema_config_check_env.runtime_data_check_env_app.CheckEnvPluginConfig import CheckEnvPluginConfig
from argrelay_schema_config_check_env.schema_config_check_env_app.CheckEnvPluginConfigSchema import (
    check_env_plugin_config_desc,
)
from argrelay_schema_config_server.runtime_data_server_plugin.PluginEntry import PluginEntry

# Standard color scheme has to be synced with `@/exe/check_env.bash`:
success_color = f"{TermColor.back_dark_green.value}{TermColor.fore_dark_black.value}"
warning_color = f"{TermColor.back_dark_yellow.value}{TermColor.fore_dark_black.value}"
failure_color = f"{TermColor.back_dark_red.value}{TermColor.fore_bright_white.value}"
field_color = TermColor.fore_bright_cyan.value
failure_message = TermColor.fore_dark_red.value
warning_message = TermColor.fore_bright_yellow.value
success_message = TermColor.fore_bright_green.value
reset_style = TermColor.reset_style.value

# Extra color scheme (available via Python code only):
offline_color = f"{TermColor.back_light_gray.value}{TermColor.fore_dark_black.value}"
offline_message = TermColor.fore_bright_magenta.value


# TODO: TODO_67_33_03_53.add_check_env_test_support.md
def main():
    # noinspection PyBroadException
    try:
        return check_env_logic()
    except BaseException as e1:
        handle_main_exception(e1)


def check_env_logic():
    argrelay_dir: str = os.path.realpath(os.path.abspath(sys.argv[1]))
    misc_helper_common.set_argrelay_dir(argrelay_dir)

    dry_run = False
    online_mode = None

    if len(sys.argv) > 3:
        raise ValueError(f"too many arguments: {' '.join(sys.argv[1:])}")
    elif len(sys.argv) == 3:
        second_arg = sys.argv[2].strip()
        if second_arg == "offline":
            online_mode = False
        elif second_arg == "online":
            online_mode = True
        elif second_arg == "dry_run":
            dry_run = True
        else:
            raise ValueError(f"unrecognized argument: {second_arg}")
    else:
        online_mode = None

    plugin_config: CheckEnvPluginConfig = check_env_plugin_config_desc.obj_from_default_file()

    total_success: bool = True

    for plugin_instance_id in plugin_config.check_env_plugin_instance_id_activate_list:
        plugin_entry: PluginEntry = plugin_config.check_env_plugin_instances[plugin_instance_id]

        if not plugin_entry.plugin_enabled:
            continue

        if (
            plugin_entry.plugin_side != PluginSide.PluginClientSideOnly
            and
            plugin_entry.plugin_side != PluginSide.PluginAnySide
        ):
            continue

        plugin_instance: PluginCheckEnvAbstract = instantiate_client_plugin(
            plugin_instance_id,
            plugin_entry,
        )
        plugin_type = plugin_instance.get_plugin_type()

        if plugin_type is PluginType.CheckEnvPlugin:
            plugin_instance: PluginCheckEnvAbstract
            plugin_instance.activate_plugin()

        if dry_run:
            continue

        check_env_results: list[CheckEnvResult] = plugin_instance.execute_check(
            online_mode,
        )
        for check_env_result in check_env_results:
            total_success: bool = print_output_line(check_env_result, total_success)

    if not total_success:
        from argrelay_lib_root.enum_desc.ClientExitCode import ClientExitCode
        exit(ClientExitCode.GeneralError.value)


def print_output_line(
    check_env_result: CheckEnvResult,
    total_success: bool,
):
    if (
        check_env_result.result_category is ResultCategory.ExecutionFailure
        or
        check_env_result.result_category is ResultCategory.VerificationFailure
    ):
        output_category = OutputCategory.is_failure
    elif check_env_result.result_category is ResultCategory.ServerOffline:
        output_category = OutputCategory.is_offline
    elif check_env_result.result_category is ResultCategory.VerificationWarning:
        output_category = OutputCategory.is_warning
    elif check_env_result.result_category is ResultCategory.VerificationSuccess:
        output_category = OutputCategory.is_success
    else:
        raise RuntimeError(check_env_result)
    total_success = total_success and not (output_category is OutputCategory.is_failure)

    # Print level:
    if output_category is OutputCategory.is_failure:
        level_color = failure_color
        level_name = "ERROR"
    elif output_category is OutputCategory.is_offline:
        level_color = offline_color
        level_name = "SKIP"
    elif output_category is OutputCategory.is_warning:
        level_color = warning_color
        level_name = "WARN"
    else:
        level_color = success_color
        level_name = "INFO"
    print(f"{level_color}{level_name}:{reset_style}", end = " ")

    # Print field:
    if check_env_result.result_key is not None:
        print(f"{field_color}{check_env_result.result_key}:{reset_style}", end = " ")
    else:
        pass

    # Print value:
    if check_env_result.result_value is not None:
        print(f"{check_env_result.result_value}", end = " ")
    else:
        pass

    # Print message:
    if check_env_result.result_message is not None:
        if output_category is OutputCategory.is_failure:
            message_color = failure_message
        elif output_category is OutputCategory.is_offline:
            message_color = offline_message
        elif output_category is OutputCategory.is_warning:
            message_color = warning_message
        else:
            message_color = success_message
        print(f"{message_color}# {check_env_result.result_message}{reset_style}", end = " ")
    else:
        pass

    # Terminate line:
    print()

    return total_success


if __name__ == "__main__":
    main()



from __future__ import annotations

from argrelay_api_plugin_server_abstract.AbstractPluginServer import AbstractPluginServer
from argrelay_app_server.relay_server.IndexModel import IndexModel
from argrelay_app_server.relay_server.QueryEngine import QueryEngine
from argrelay_lib_root.enum_desc.PluginType import PluginType
from argrelay_schema_config_server.runtime_data_server_app.EnvelopeCollection import EnvelopeCollection


class AbstractLoader(AbstractPluginServer):
    """
    `LoaderPlugin` implements loading of `data_envelope` (which are later selected on the CLI via args).

    This is a way to load data once on server start.

    The source of data can be anything (hence, it is a plugin).
    """

    def get_plugin_type(
        self,
    ) -> PluginType:
        return PluginType.LoaderPlugin

    def list_index_models(
        self,
    ) -> list[IndexModel]:
        """
        Provides list of index models for FS_45_08_22_15 index model API.

        TODO: TODO_23_17_31_04: publish `index_model` by delegator plugins.
        """
        raise NotImplementedError

    def load_envelope_collections(
        self,
        query_engine: QueryEngine,
    ) -> list[EnvelopeCollection]:
        """
        Return loaded `envelope_collection`-s possibly joining them with
        data already stored in the backend accessible via `query_engine`.
        """
        return []



from enum import (
    auto,
    Enum,
)


class FuncState(Enum):
    """
    State of the func (its readiness).
    """

    fs_unplugged = auto()
    """
    Used for func that are not plugged anywhere in the tree (but they are published).

    See `DelegatorAbstract.get_supported_func_envelopes`.
    """

    fs_ignorable = auto()
    """
    Not `fs_unplugged`, but still not very useful.
    """

    fs_demo = auto()
    """
    Not meant to work - for `argrelay` demo only.
    """

    fs_alpha = auto()
    """
    In "alpha" testing.
    """

    fs_beta = auto()
    """
    In "beta" testing.
    """

    fs_gamma = auto()
    """
    In GA (general availability).
    """

    def __str__(self):
        return self.name



from __future__ import annotations

from argrelay_api_plugin_server_abstract.DelegatorAbstract import get_func_id_from_invocation_input
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_server_plugin_demo.demo_git.DelegatorGitRepoBase import DelegatorGitRepoBase
from argrelay_lib_server_plugin_demo.demo_git.GitRepoEnvelopeClass import GitRepoEnvelopeClass
from argrelay_lib_server_plugin_demo.demo_git.GitRepoPropName import GitRepoPropName
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
)
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    func_id_,
    search_control_list_,
)
from argrelay_schema_config_server.schema_config_interp.SearchControlSchema import populate_search_control

func_id_desc_git_commit_ = "func_id_desc_git_commit"

repo_container_ipos_ = 1


class DelegatorGitRepoDescCommit(DelegatorGitRepoBase):
    """
    Implements `func_id_desc_git_commit_` for FS_67_16_61_97 git_plugin.
    """

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        commit_search_control = populate_search_control(
            GitRepoEnvelopeClass.class_git_commit.name,
            {
                ReservedPropName.envelope_class.name: GitRepoEnvelopeClass.class_git_commit.name,
            },
            [
                # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
                {"class": ReservedPropName.envelope_class.name},

                {"category": GitRepoPropName.git_repo_object_category.name},

                {"alias": GitRepoPropName.git_repo_alias.name},
                {"content": GitRepoPropName.git_repo_content_type.name},
                {"name": GitRepoPropName.git_repo_root_base_name.name},
                {"path": GitRepoPropName.git_repo_root_rel_path.name},
                {"base": GitRepoPropName.git_repo_root_abs_path.name},

                {"date": GitRepoPropName.git_repo_commit_date.name},
                {"email": GitRepoPropName.git_repo_commit_author_email.name},
                {"time": GitRepoPropName.git_repo_commit_time.name},
                {"hex": GitRepoPropName.git_repo_short_commit_id.name},
            ],
        )

        func_envelopes = []

        given_function_envelope = {
            instance_data_: {
                func_id_: func_id_desc_git_commit_,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                    commit_search_control,
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
            ReservedPropName.help_hint.name: "Describe Git commit",
            ReservedPropName.func_state.name: FuncState.fs_demo.name,
            ReservedPropName.func_id.name: func_id_desc_git_commit_,
        }
        func_envelopes.append(given_function_envelope)

        return func_envelopes

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        assert get_func_id_from_invocation_input(invocation_input) == func_id_desc_git_commit_
        raise RuntimeError("ERROR: not implemented for demo")



from typing import Union

from argrelay_api_plugin_server_abstract.AbstractPluginServer import AbstractPluginServer
from argrelay_lib_root.enum_desc.PluginType import PluginType


class ConfiguratorAbstract(AbstractPluginServer):
    """
    `PluginType.ConfiguratorPlugin` implements logic to configure `argrelay` server
    when static config is not good enough.
    """

    def get_plugin_type(
        self,
    ) -> PluginType:
        return PluginType.ConfiguratorPlugin

    def provide_project_title(
        self,
    ) -> Union[str, None]:
        """
        Names an instance of `argrelay`.
        """
        return None

    def provide_project_page_url(
        self,
    ) -> Union[str, None]:
        """
        Project a URL to project page.
        """
        return None

    def provide_project_git_commit_display_string(
        self,
    ) -> Union[str, None]:
        """
        Returns commit id (in any format) to be shown as is.

        This commit display string is used as text for a link (purely display function).

        The link itself is generated by concatenating two components - see:
        *   `provide_project_commit_id_url_prefix`
        *   `provide_project_git_commit_id`
        """
        return None

    def provide_project_git_files_by_commit_id_url_prefix(
        self,
    ) -> Union[str, None]:
        """
        Returns a URL prefix to form valid URL by concatenation of:
        *   `provide_project_git_files_by_commit_id_url_prefix`
        *   `provide_project_git_commit_id`
        *   `provide_project_git_repo_relative_argrelay_dir`
        *   `provide_project_current_config_path`

        For example:
        https://github.com/argrelay/argrelay/tree/
        to be used for:
        https://github.com/argrelay/argrelay/tree/ba1e6c684c476558181a8e127d9c34f1feaf5cde/dst/.github
        """
        return None

    def provide_project_commit_id_url_prefix(
        self,
    ) -> Union[str, None]:
        """
        Returns a URL prefix to form valid URL by concatenation with:
        *   `provide_project_git_commit_id`

        The display text for the link is taken from `provide_project_git_commit_display_string` instead.

        For example:
        https://github.com/argrelay/argrelay/commit/
        to be used for:
        https://github.com/argrelay/argrelay/commit/70bf34b3f773b576c8153ada9ea3baa5b228b8a0
        """
        return None

    def provide_project_git_commit_id(
        self,
    ) -> Union[str, None]:
        """
        Returns commit id in specific format (e.g. full) which should work as part of a URL.

        It is concatenated with `provide_project_git_commit_id_url_prefix` to form a URL.

        The display text for the link is taken from `provide_project_git_commit_display_string` instead.
        """
        return None

    def provide_project_git_commit_time(
        self,
    ) -> Union[int, None]:
        """
        Return commit time in seconds since epoch (Unix time).

        The commit time should normally be taken for the same commit id as `provide_project_git_commit_id`.
        """
        return None

    def provide_project_git_repo_relative_argrelay_dir(
        self,
    ) -> Union[str, None]:
        """
        Return relative path of `argrelay_dir` within project git repo.

        It is used to provide link to config files by concatenation of:
        *   `provide_project_git_files_by_commit_id_url_prefix`
        *   `provide_project_git_commit_id`
        *   `provide_project_git_repo_relative_argrelay_dir`
        *   `provide_project_current_config_path`
        """
        return None

    def provide_project_current_config_path(
        self,
    ) -> Union[str, None]:
        """
        Returns path where `@/conf/` symlink points to (or `conf` str directly if `@/conf/` is a dir).

        The path is relative to `argrelay_dir` (`@/`).

        The purpose is to compose a valid URL concatenating it with other parts.
        For this to work, `@/conf/` should be a symlink with relative path (pointing inside `argrelay_dir`).

        It is used to provide link to config files by concatenation of:
        *   `provide_project_git_files_by_commit_id_url_prefix`
        *   `provide_project_git_commit_id`
        *   `provide_project_git_repo_relative_argrelay_dir`
        *   `provide_project_current_config_path`
        """
        return None



from copy import deepcopy

from marshmallow import (
    fields,
    RAISE,
)

from argrelay_api_server_cli.schema_request.CallContextSchema import call_context_desc
from argrelay_api_server_cli.schema_response.InterpResultSchema import (
    interp_result_desc,
    InterpResultSchema,
)
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_lib_server_plugin_core.plugin_delegator.SchemaCustomDataDelegatorError import (
    error_delegator_custom_data_desc,
)
from argrelay_schema_config_server.schema_config_server_plugin.PluginEntrySchema import plugin_entry_desc

call_ctx_ = "call_ctx"
delegator_plugin_entry_ = "delegator_plugin_entry"
custom_plugin_data_ = "custom_plugin_data"


class InvocationInputSchema(InterpResultSchema):
    class Meta:
        unknown = RAISE
        ordered = True

    model_class = InvocationInput

    call_ctx = fields.Nested(
        call_context_desc.dict_schema,
        required = True,
    )

    delegator_plugin_entry = fields.Nested(
        plugin_entry_desc.dict_schema,
        required = True,
    )

    custom_plugin_data = fields.Dict(
        required = False,
    )


_call_ctx_example = deepcopy(call_context_desc.dict_example)
_invocation_input_example = deepcopy(interp_result_desc.dict_example)
_invocation_input_example.update({
    call_ctx_: _call_ctx_example,
    delegator_plugin_entry_: plugin_entry_desc.dict_example,
    custom_plugin_data_: error_delegator_custom_data_desc.dict_example,
})

invocation_input_desc = TypeDesc(
    dict_schema = InvocationInputSchema(),
    ref_name = InvocationInputSchema.__name__,
    dict_example = _invocation_input_example,
    default_file_path = "",
)



from __future__ import annotations

from argrelay_api_plugin_server_abstract.delegator_utils import redirect_to_not_disambiguated_error
from argrelay_api_plugin_server_abstract.DelegatorAbstract import get_func_id_from_invocation_input
from argrelay_api_plugin_server_abstract.DelegatorJumpAbstract import DelegatorJumpAbstract
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import (
    function_container_ipos_,
    InterpContext,
)
from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_root.enum_desc.SpecialFunc import SpecialFunc
from argrelay_lib_root.enum_desc.TermColor import TermColor
from argrelay_lib_server_plugin_core.plugin_delegator.client_invocation_utils import prohibit_unconsumed_args
from argrelay_lib_server_plugin_core.plugin_interp.FuncTreeInterpFactory import tree_step_prop_name_prefix_
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    func_id_,
    search_control_list_,
)
from argrelay_schema_config_server.schema_config_interp.SearchControlSchema import search_control_desc

subsequent_function_container_ipos_ = function_container_ipos_ + 1


class DelegatorHelp(DelegatorJumpAbstract):

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        func_envelopes = [{
            instance_data_: {
                func_id_: SpecialFunc.func_id_help_hint.name,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
            ReservedPropName.help_hint.name: "List defined function matching search criteria with their help hints",
            ReservedPropName.func_state.name: FuncState.fs_gamma.name,
            ReservedPropName.func_id.name: SpecialFunc.func_id_help_hint.name,
        }]
        return func_envelopes

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        if interp_ctx.curr_container_ipos >= subsequent_function_container_ipos_:
            subsequent_function_container = interp_ctx.envelope_containers[(
                subsequent_function_container_ipos_
            )]
            subsequent_function_container.data_envelopes = (
                local_server
                .get_query_engine()
                .query_data_envelopes_for(subsequent_function_container)
            )

            delegator_plugin_instance_id = self.plugin_instance_id

            custom_plugin_data = search_control_desc.dict_schema.dump(subsequent_function_container.search_control)

            invocation_input = InvocationInput.with_interp_context(
                interp_ctx,
                delegator_plugin_entry = local_server.plugin_config.server_plugin_instances[
                    delegator_plugin_instance_id
                ],
                custom_plugin_data = custom_plugin_data,
            )
            return invocation_input
        else:
            return redirect_to_not_disambiguated_error(
                interp_ctx,
                local_server.plugin_config,
                ReservedEnvelopeClass.class_function.name,
            )

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        if get_func_id_from_invocation_input(invocation_input) == SpecialFunc.func_id_help_hint.name:

            prohibit_unconsumed_args(invocation_input)

            for data_envelope in (
                invocation_input
                    .envelope_containers[subsequent_function_container_ipos_]
                    .data_envelopes
            ):

                # Dynamic `search_control` does not contain full paths to funcs all the time -
                # it is based on curr path within FS_01_89_09_24 interp tree and FS_26_43_73_72 func tree.
                # Recognize props with the prefix instead (which are set on loading and remain static):
                tree_step_prop_names = []
                for prop_name in data_envelope:
                    if prop_name.startswith(tree_step_prop_name_prefix_):
                        tree_step_prop_names.append(prop_name)
                # Iterate through sorted prop names - `tree_step_prop_name_prefix_` is numbered:
                tree_step_prop_names.sort()
                for tree_step_prop_name in tree_step_prop_names:
                    print(f"{data_envelope[tree_step_prop_name]}", end = " ")

                # TODO: FS_02_25_41_81 (func_id_query_enum_items): perform color control only if the output is a terminal:

                print(TermColor.known_envelope_id.value, end = "")
                print("#", end = " ")
                print(TermColor.reset_style.value, end = "")

                print(TermColor.known_envelope_id.value, end = "")
                print(f"{data_envelope[instance_data_][func_id_]}", end = " ")
                print(TermColor.reset_style.value, end = "")

                print(TermColor.help_hint.value, end = "")
                print("#", end = " ")
                print(TermColor.reset_style.value, end = "")

                if ReservedPropName.help_hint.name in data_envelope:
                    print(TermColor.help_hint.value, end = "")
                    print(
                        f"{data_envelope[ReservedPropName.help_hint.name]}",
                        end = " ",
                    )
                    print(TermColor.reset_style.value, end = "")
                else:
                    print(TermColor.no_help_hint.value, end = "")
                    print(f"[no `{ReservedPropName.help_hint.name}`]", end = " ")
                    print(TermColor.reset_style.value, end = "")

                print()



from dataclasses import (
    dataclass,
    field,
)


@dataclass
class GuiBannerConfig:
    meter_html: str = field()
    tagline_html: str = field()
    header_html: str = field()
    footer_html: str = field()



from __future__ import annotations

from marshmallow import (
    fields,
    RAISE,
    validates_schema,
    ValidationError,
)

from argrelay_app_server.runtime_context.SearchControl import SearchControl
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc

collection_name_ = "collection_name"

props_to_values_dict_ = "props_to_values_dict"
"""
Dict of `prop_name` assigned to their `prop_value`-s.
"""

arg_name_to_prop_name_map_ = "arg_name_to_prop_name_map"
"""
List of `arg_name`-s to use for `dictated_arg`-s during interpretation and `prop_name`-s to use in query.
"""


class SearchControlSchema(ObjectSchema):
    """
    Implements:
    *   FS_46_96_59_05 `init_control`
    *   FS_31_70_49_15 `search_control`
    """

    class Meta:
        unknown = RAISE
        strict = True

    model_class = SearchControl

    collection_name = fields.String(
        required = True,
    )

    # Specifies:
    # *   FS_46_96_59_05 `init_control`: what `prop_value` should be set for `prop_name`.
    props_to_values_dict = fields.Dict(
        # `prop_name`:
        keys = fields.String(),
        # `prop_value`:
        values = fields.String(),
        required = True,
    )

    # Specifies:
    # *   What `prop_name`-s will be used in search of the `data_envelope`.
    # *   How they are mapped to `arg_name`-s - see:
    #     *   FS_10_93_78_10: `arg_name_to_prop_name_map`
    #     *   FS_20_88_05_60: `dictated_arg`-s: maps `arg_name` (in command line) to `prop_name` (in `data_envelope`).
    # Each `list` item must contain singleton `dict` (single pair of mapping from `arg_name` to `prop_name`).
    arg_name_to_prop_name_map = fields.List(
        fields.Dict(
            # `arg_name`:
            keys = fields.String(),
            # `prop_name`:
            values = fields.String(),
            required = True,
        ),
        required = True,
    )

    @validates_schema
    def validate_known(
        self,
        input_dict: dict,
        **kwargs,
    ):
        for arg_name_to_prop_name_entry in input_dict[arg_name_to_prop_name_map_]:
            # ensure there is only one `arg_name` per `arg_name_to_prop_name_entry`:
            if len(arg_name_to_prop_name_entry) != 1:
                raise ValidationError("only one key per dict is allowed")


search_control_desc = TypeDesc(
    dict_schema = SearchControlSchema(),
    ref_name = SearchControlSchema.__name__,
    dict_example = {
        collection_name_: ReservedEnvelopeClass.class_function.name,
        props_to_values_dict_: {
            "prop_name_a": "prop_value_a",
            "prop_name_b": "prop_value_b",
        },
        arg_name_to_prop_name_map_: [
            {
                "arg_name_a": "prop_name_a",
            },
            {
                "arg_name_b": "prop_name_b",
            },
        ],
    },
    default_file_path = "",
)


def populate_search_control(
    collection_name: str,
    props_to_values_dict: dict[str, str],
    arg_name_to_prop_name_map: list[dict],
) -> dict:
    search_control: dict = {
        collection_name_: collection_name,
        props_to_values_dict_: props_to_values_dict,
        arg_name_to_prop_name_map_: arg_name_to_prop_name_map,
    }

    return search_control



from __future__ import annotations

from argrelay_lib_server_plugin_core.plugin_delegator.DelegatorNoopBase import DelegatorNoopBase


class DelegatorNoopGroup(DelegatorNoopBase):
    """
    The purpose of this delegator is to group other delegators.

    It groups delegators as its dependency list - see `plugin_dependencies`.
    """
